#include "cnfsolver.h"
#include "planner.h"

/*************************************************************************************/
#ifdef _MSC_VER
#include <ctime>

static inline double cpuTime(void) {
    return (double)clock() / CLOCKS_PER_SEC; }
#else

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

static inline double cpuTime(void) {
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_utime.tv_sec + (double)ru.ru_utime.tv_usec / 1000000; }
#endif


#if defined(__linux__)
static inline int memReadStat(int field)
{
    char    name[256];
    pid_t pid = getpid();
    sprintf(name, "/proc/%d/statm", pid);
    FILE*   in = fopen(name, "rb");
    if (in == NULL) return 0;
    int     value;
    for (; field >= 0; field--)
        fscanf(in, "%d", &value);
    fclose(in);
    return value;
}
static inline uint64_t memUsed() { return (uint64_t)memReadStat(0) * (uint64_t)getpagesize(); }


#elif defined(__FreeBSD__)
static inline uint64_t memUsed(void) {
    struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return ru.ru_maxrss*1024; }


#else
static inline uint64_t memUsed() { return 0; }
#endif

#if defined(__linux__)
#include <fpu_control.h>
#endif

//=================================================================================================
// DIMACS Parser:


template<class B>
static void skipWhitespace(B& in) {
    while ((*in >= 9 && *in <= 13) || *in == 32)
        ++in; }

template<class B>
static void skipLine(B& in) {
    for (;;){
        if (*in == EOF || *in == '\0') return;
        if (*in == '\n') { ++in; return; }
        ++in; } }

template<class B>
static int parseInt(B& in) {
    int     val = 0;
    bool    neg = false;
    skipWhitespace(in);
    if      (*in == '-') neg = true, ++in;
    else if (*in == '+') ++in;
    if (*in < '0' || *in > '9') reportf("PARSE ERROR! Unexpected char: %c\n", *in), exit(3);
    while (*in >= '0' && *in <= '9')
        val = val*10 + (*in - '0'),
        ++in;
    return neg ? -val : val; }

template<class B>
static void readClause(B& in, Solver& S, vec<Lit>& lits) {
    int     parsed_lit, var;
    lits.clear();
    for (;;){
        parsed_lit = parseInt(in);
        if (parsed_lit == 0) break;
        var = abs(parsed_lit)-1;
        while (var >= S.nVars()) S.newVar();
        lits.push( (parsed_lit > 0) ? Lit(var) : ~Lit(var) );
    }
}

template<class B>
static bool match(B& in, char* str) {
    for (; *str != 0; ++str, ++in)
        if (*str != *in)
            return false;
    return true;
}


template<class B>
static void parse_DIMACS_main(B& in, Solver& S) {
    vec<Lit> lits;
    for (;;){
        skipWhitespace(in);
        if (*in == EOF)
            break;
        else if (*in == 'p'){
            if (match(in, "p cnf")){
                int vars    = parseInt(in);
                int clauses = parseInt(in);
                reportf("|  Number of variables:  %-12d                                         |\n", vars);
                reportf("|  Number of clauses:    %-12d                                         |\n", clauses);
            }else{
                reportf("PARSE ERROR! Unexpected char: %c\n", *in), exit(3);
            }
        } else if (*in == 'c' || *in == 'p')
            skipLine(in);
        else
            readClause(in, S, lits),
            S.addClause(lits);
    }
}

// Inserts problem into solver.
//
static void parse_DIMACS(gzFile input_stream, Solver& S) {
    //StreamBuffer in(input_stream);
    //parse_DIMACS_main(in, S); 
}

//=================================================================================================


void printStats(Solver& solver)
{
    double   cpu_time = cpuTime();
    uint64_t mem_used = memUsed();
    reportf("restarts              : %lld\n", solver.starts);
    reportf("conflicts             : %-12lld   (%.0f /sec)\n", solver.conflicts   , solver.conflicts   /cpu_time);
    reportf("decisions             : %-12lld   (%4.2f %% random) (%.0f /sec)\n", solver.decisions, (float)solver.rnd_decisions*100 / (float)solver.decisions, solver.decisions   /cpu_time);
    reportf("propagations          : %-12lld   (%.0f /sec)\n", solver.propagations, solver.propagations/cpu_time);
    reportf("conflict literals     : %-12lld   (%4.2f %% deleted)\n", solver.tot_literals, (solver.max_literals - solver.tot_literals)*100 / (double)solver.max_literals);
    if (mem_used != 0) reportf("Memory used           : %.2f MB\n", mem_used / 1048576.0);
    reportf("CPU time              : %g s\n", cpu_time);
}

//Solver* solver;

static void SIGINT_handler(int signum) {
    reportf("\n"); reportf("*** INTERRUPTED ***\n");
    //printStats(*solver);
    reportf("\n"); reportf("*** INTERRUPTED ***\n");
    exit(1); }


//=================================================================================================
// Main:

void printUsage(char** argv)
{
    reportf("USAGE: %s [options] <input-file> <result-output-file>\n\n  where input may be either in plain or gzipped DIMACS.\n\n", argv[0]);
    reportf("OPTIONS:\n\n");
    reportf("  -polarity-mode = {true,false,rnd}\n");
    reportf("  -decay         = <num> [ 0 - 1 ]\n");
    reportf("  -rnd-freq      = <num> [ 0 - 1 ]\n");
    reportf("  -verbosity     = {0,1,2}\n");
    reportf("\n");
}


const char* hasPrefix(const char* str, const char* prefix)
{
    int len = strlen(prefix);
    if (strncmp(str, prefix, len) == 0)
        return str + len;
    else
        return NULL;
}


int main0(int argc, char** argv)
{
    Solver      S;
    S.verbosity = 1;


    int         i;
        double cpu_time = cpuTime();

    //solver = &S;
    signal(SIGINT,SIGINT_handler);
    signal(SIGHUP,SIGINT_handler);

    if (argc == 1)
        reportf("Reading from standard input... Use '-h' or '--help' for help.\n");

    gzFile in = (argc == 1) ? gzdopen(0, "rb") : gzopen(argv[1], "rb");
    if (in == NULL)
        reportf("ERROR! Could not open file: %s\n", argc == 1 ? "<stdin>" : argv[1]), exit(1);

    reportf("============================[ Problem Statistics ]=============================\n");
    reportf("|                                                                             |\n");

    parse_DIMACS(in, S);
    gzclose(in);
    FILE* res = (argc >= 3) ? fopen(argv[2], "wb") : NULL;

    double parse_time = cpuTime() - cpu_time;
    reportf("|  Parsing time:         %-12.2f s                                       |\n", parse_time);

    if (!S.simplify()){
        reportf("Solved by unit propagation\n");
        if (res != NULL) fprintf(res, "UNSAT\n"), fclose(res);
        printf("UNSATISFIABLE\n");
        exit(20);
    }

    bool ret = S.solve();
    printStats(S);
    reportf("\n");
    printf(ret ? "SATISFIABLE\n" : "UNSATISFIABLE\n");
    if (res != NULL){
        if (ret){
            fprintf(res, "SAT\n");
            for (int i = 0; i < S.nVars(); i++)
                if (S.model[i] != l_Undef)
                    fprintf(res, "%s%s%d", (i==0)?"":" ", (S.model[i]==l_True)?"":"-", i+1);
            fprintf(res, " 0\n");
        }else
            fprintf(res, "UNSAT\n");
        fclose(res);
    }

#ifdef NDEBUG
    exit(ret ? 10 : 20);     // (faster than "return", which will invoke the destructor for 'Solver')
#endif
}

//=======================Minisat SAT Solver=======================================================================

void readClause1(const Literals& x, Solver& S, vec<Lit>& lits) {
    int var;
	Literals::const_iterator it;
    lits.clear();
	for (it = x.begin(); it != x.end(); it++) {
		var = (*it)/2;                
        while (var >= S.nVars()) S.newVar();
        lits.push( (*it % 2 == 0) ? Lit(var) : ~Lit(var) );
    }
}

void readClause1(const Literals& x, Solver& S, vec<Lit>& lits,const boost::dynamic_bitset<>& m_L) {
    int var;
	Literal l;
	Literals::const_iterator it;
    lits.clear();
	for (it = x.begin(); it != x.end(); it++) {
		NEGATE(*it, l);
		if (m_L[l])
			continue;
		var = (*it)/2;                
        while (var >= S.nVars()) S.newVar();
        lits.push( (*it % 2 == 0) ? Lit(var) : ~Lit(var) );
    }
}


void readUnitClause(const Literal& l, Solver& S, vec<Lit>& lits) {
    int var;	
    lits.clear();
	
	var = l/2;
    while (var >= S.nVars()) S.newVar();
    lits.push( (l % 2 == 0) ? Lit(var) : ~Lit(var) );    
}



//=======================Revival CNF Simplification=======================================================================

void readClause(const Literals& x, veci* lits) {
    int var;
	Literals::const_iterator it;
    veci_resize(lits,0);

	for (it = x.begin(); it != x.end(); it++) {
		var = (*it)/2;                
        veci_push(lits, ( *it % 2 == 0 ? tolit(var) : lit_neg(tolit(var))));        
    }	
}

void readUnitClause(const Literal& x, veci* lits) {
    int var;	
    veci_resize(lits,0);	
	var = x/2;                
    veci_push(lits, ( x % 2 == 0 ? tolit(var) : lit_neg(tolit(var))));
}


bool parse_CNFtoSimp(CState* cs, solver* s) {
	Literals::const_iterator it;
	Literal l;
	set<CSClause*>::const_iterator its;
	veci lits;
    veci_new(&lits);
	int i =0;
	//Parse set of literals L
	//cout << "0"; fflush(stdout);	
	for ( l = cs->m_L->m_L.find_first(); l < cs->m_L->m_L.size(); ) {			
		lit* begin;
		readUnitClause(l, &lits);		
		begin = veci_begin(&lits);
        if (!solver_addclause(s, begin, begin+veci_size(&lits))){			
            veci_delete(&lits);			
            return false;
		}
		l = cs->m_L->m_L.find_next(l);
	}	
	
	//Parse set of sets of literals S
	for (its = cs->m_clauses.begin(); its != cs->m_clauses.end(); its++) {		
		lit* begin;
		readClause((*its)->m_literals, &lits);
		begin = veci_begin(&lits);
        if (!solver_addclause(s, begin, begin+veci_size(&lits))){
            veci_delete(&lits);
            return false;
		}
	}
	
	veci_delete(&lits);		
	return (solver_simplify(s));
}

bool parse_CNFtoSimp(set<Literals>* cnf, solver* s) {	
	set<Literals>::const_iterator its;
	veci lits;
    veci_new(&lits);	
	
	//Parse set of sets of literals S
	for (its = cnf->begin(); its != cnf->end(); its++) {
		lit* begin;
		readClause(*its, &lits);
		begin = veci_begin(&lits);
        if (!solver_addclause(s, begin, begin+veci_size(&lits))){
            veci_delete(&lits);
            return false;
		}
	}

	veci_delete(&lits);
    return solver_simplify(s);
}

bool simplifyCNF(CState* cs){
  
  cs->m_planner->m_timer->start(SIMPLIFICATION_TIMER);

  solver* s = solver_new();  
  FILE *  in;
  veci    learnt_clause;
  clause *currentClause,*learnt_cl=NULL,*c;  
  lit  *begin, *end;
  int j,k,l,lit,noConflict,change=0,useF1,actualCurrentNumClause,
    nbIteration=0,my_lit,num_assuming,p1,p2;
  unsigned int size;

  int i,globalModif=0,lastModif=0,timeOut=0,num_unit=0, refresh;
  //int i,globalModif,lastModif,timeOut=0,num_unit=0;
  
#ifdef RANDOMIZED_ALGORITHM
  srand(time(NULL));
#else
  srand(1);
#endif
  //cout << "0"; fflush(stdout);  
  /* Parse CNF and initialize variables */  
  if (cs->m_clauses.size() == 0) {
	  solver_delete(s);
	  return true;
  }
  //cout << "1"; fflush(stdout);	
  if (!parse_CNFtoSimp(cs, s))	 {
	  solver_delete(s);
	  return true; //CNF is unsatisfiable
  }    
  //cout << "2"; fflush(stdout);	
  int lUnitclauses = 0, lUnitclauses_new = 0;
  int lclauses=0, lclauses_new;
  int llits=0, llits_new=0;
  

  for (i=0,size=s->size;i<size;i++){
    if(s->assigns[i]!=lB_Undef){
      lUnitclauses_new++;
    }      
  }

  lclauses_new = s->stats.clauses;
  for(i=0; i<lclauses_new; i++){
    currentClause= (clause*) (vecp_begin(&(s->clauses)))[i];
    if(currentClause){
      llits_new += clause_size(currentClause);       
    }
  }
  
  while (lUnitclauses != lUnitclauses_new || lclauses_new != lclauses || llits_new != llits) {  
  
	actualCurrentNumClause=lastModif=s->stats.clauses;
	globalModif=1;
  
  /* The flag globalModif is true if literals have been removed from at least one clause.
     Hence, new propagations could be possible, and new subclauses could be produced. 
     All clauses are tested as long as this flag is true (except if a time out is defined and is reached) */
	while(globalModif){
		++nbIteration;	
		globalModif=0;
		for(i=0;i<(globalModif?s->stats.clauses:lastModif);i=((change)?(i):(i+1))){		  
		  if(!(currentClause=(clause*) (vecp_begin(&(s->clauses)))[i])) {change=0; continue;}
		  /* The considered clause is removed from the CNF */
		  simulate_removeClause(s, (vecp_begin(&(s->clauses)))[i]);
		  change=0,noConflict=1;
		  if(solver_propagate(s)) {
			  solver_delete(s);
			  return true;
		  }

		  /* Update the clause w.r.t. the current assignment */
		  //if(!(refreshClause(s,&currentClause))){			
		  if(!(refresh = refreshClause(s,&currentClause))){
			free_clause(currentClause);
			actualCurrentNumClause--;
			continue;
		  }
		  else if (refresh == -1) return false;

		  size=clause_size(currentClause);
		  num_assuming=0;
		  /* The MOMS-like sort of the clause has been removed */
		  /* MOMS_sort(currentClause); */
		  /* Two randomly selected literals are reversed */
		  p1=random()%(clause_size(currentClause));
		  p2=random()%(clause_size(currentClause));
		  reverseLit(currentClause,p1,p2);
		  /* Assignment of the opposite of each literal from the clause */
		  for(j=0;(j<size) && (noConflict) && (!change);j++){	  
			if(clause_size(currentClause)==1) {break;}
			lit=clause_giveLit(currentClause,j);
			num_assuming++;
			assume(s,negation(lit)); 
			noConflict=((learnt_cl=solver_propagate(s))==NULL);	
			/* The following line can be safely uncommented 
			   (provided that '}' is added just before the "if(!noConflict){" line). 
			   This would remove the next loop in case of conflict. 
			   However, performing this test appears to degrade performances. (weird...) */
			//if(noConflict){
			for(k=j+1;k<size;k++){
			  my_lit=clause_giveLit(currentClause,k);
			  // The literal is assigned to the same value. the remaining part of the clause can be "omitted" 
			  if(s->assigns[lit_var(my_lit)]!=lB_Undef){
				if(AmISat(my_lit)){
				  if(clause_size(currentClause)!=num_assuming+1){
				change=1; 
				clause_begin(currentClause)[num_assuming]=my_lit;
				currentClause->size_learnt=(num_assuming+1)<<1;
				  }else{
				// In this case, the clause is redundant but no shorter clause can be deduced 
				// (last literal already assigned) 
				j=size;
				break;
				  }
				}
				// The literal is assigned to the opposite value. it can be removed from the clause 
				else{ 
				  for(l=k+1;l<size;l++) 
				clause_begin(currentClause)[l-1]=clause_giveLit(currentClause,l); 
				  currentClause->size_learnt=(size-1)<<1; 
				  change=1;
				}
				break;
			  }
			}
		  }
		  if(!noConflict){
			if(clause_size(currentClause)!=num_assuming){
				change=1; 
				currentClause->size_learnt=(num_assuming)<<1;
			}
			veci_new(&learnt_clause);
			veci_resize(&learnt_clause,0);
			solver_analyze(s, learnt_cl, &learnt_clause);
			begin = veci_begin(&learnt_clause);
			end   = begin + veci_size(&learnt_clause);
			if(veci_size(&learnt_clause) > 1){
				#ifndef DONT_LEARN_ANYTHING
					  if(veci_size(&learnt_clause) < clause_size(currentClause)){
						c     = clause_new(s,begin,end,1);
						vecp_push(&s->learnts,c);
						s->stats.learnts++;
					  }
				#endif
			}
			else {
			  solver_canceluntil(s,s->root_level);
			  num_unit++;
			  enqueue(s,*begin,0);
			}
			if(clause_size(currentClause)!=num_assuming){
				currentClause->size_learnt=(num_assuming)<<1,change=1;
			}
		  }
		  solver_canceluntil(s,s->root_level);
		  if(clause_size(currentClause)==1){			
			num_unit++;
			enqueue(s,clause_giveLit(currentClause,0),(clause*)0); 
			free_clause(currentClause);
			actualCurrentNumClause--;
			continue; 
		  }
		  simulate_addItAgain(s,(vecp_begin(&(s->clauses)))[i],currentClause);
		  if(change){
			if(!timeOut) {globalModif=1;  lastModif=i+1;} 			
		  }
		}
	}    
  
	if(solver_propagate(s)) {
		solver_delete(s);
		return true;
	}

	solver_canceluntil(s,s->root_level);
  
	//fprintf(stderr,"#ITERATIONS %d\n",nbIteration);
	//fprintf(stderr,"#LEARNT CLAUSES %d\n",(int) s->stats.learnts);
  
	/* Export produced CNF */
	/* Detected unit clauses first, then shortened clauses from the original CNF, then learnt clauses (3 loops) */
	/*  
	printf("p cnf %d %d\n",s->size,(int) ((actualCurrentNumClause)+((int) s->stats.learnts)+countExportedUnit(s)));
	for(i=0,size=s->size;i<size;i++){
		if(s->assigns[i]!=lB_Undef){
		  printf("%d 0\n",s->assigns[i]==lB_True?i+1:-(i+1));
		}      
	}
	

	for(i=0,size=s->stats.clauses;i<size;i++){
		currentClause= (clause*) (vecp_begin(&(s->clauses)))[i];
		if(currentClause){
		  k=clause_size(currentClause);
		  for(j=0;j<k;j++) printf("%d ",wellWritten(clause_giveLit(currentClause,j)));
		  printf("0\n");
		}
	}  
	*/

	//Compute number of clauses and total length to determine	further simplification
	lUnitclauses = lUnitclauses_new; 
	lclauses = lclauses_new;
	llits = llits_new;
	lUnitclauses_new = 0;
	for(i=0,size=s->size;i<size;i++){
		if(s->assigns[i]!=lB_Undef){
		  lUnitclauses_new++;
		}      
	}

	lclauses_new = s->stats.clauses;
	llits_new = 0;
	for(i=0; i<lclauses_new; i++){
		currentClause= (clause*) (vecp_begin(&(s->clauses)))[i];
		if(currentClause){
		  llits_new += clause_size(currentClause);       
		}
	}    

  } /* My while */
  //cout << "3"; fflush(stdout);

  cs->m_planner->m_timer->end(SIMPLIFICATION_TIMER);
  CNF_extract(cs, s, lUnitclauses_new, lclauses_new, llits_new);
  return true;
}


bool simplifyCNF(set<Literals>* cs){
  solver* s = solver_new();  
  FILE *  in;
  veci    learnt_clause;
  clause *currentClause,*learnt_cl=NULL,*c;  
  lit  *begin, *end;
  int j,k,l,lit,noConflict,change=0,useF1,actualCurrentNumClause,
    nbIteration=0,my_lit,num_assuming,p1,p2;
  unsigned int size;

  int i,globalModif=0,lastModif=0,timeOut=0,num_unit=0, refresh;
  
#ifdef RANDOMIZED_ALGORITHM
  srand(time(NULL));
#else
  srand(1);
#endif
  
  /* Parse CNF and initialize variables */  
  if (!parse_CNFtoSimp(cs, s)) {
	  solver_delete(s);
	  return true; //CNF is unsatisfiable
  }
   
  int lUnitclauses = 0, lUnitclauses_new = 0;
  int lclauses=0, lclauses_new;
  int llits=0, llits_new=0;

  for(i=0,size=s->size;i<size;i++){
    if(s->assigns[i]!=lB_Undef){
      lUnitclauses_new++;
    }      
  }

  lclauses_new = s->stats.clauses;
  for(i=0; i<lclauses_new; i++){
    currentClause= (clause*) (vecp_begin(&(s->clauses)))[i];
    if(currentClause){
      llits_new += clause_size(currentClause);       
    }
  }

  while (lUnitclauses != lUnitclauses_new || lclauses_new != lclauses || llits_new != llits) {  
  
	actualCurrentNumClause=lastModif=s->stats.clauses;
	globalModif=1;
  
  /* The flag globalModif is true if literals have been removed from at least one clause.
     Hence, new propagations could be possible, and new subclauses could be produced. 
     All clauses are tested as long as this flag is true (except if a time out is defined and is reached) */
	while(globalModif){
		++nbIteration;	
		globalModif=0;
		for(i=0;i<(globalModif?s->stats.clauses:lastModif);i=((change)?(i):(i+1))){		  
		  if(!(currentClause=(clause*) (vecp_begin(&(s->clauses)))[i])) {change=0; continue;}
		  /* The considered clause is removed from the CNF */
		  simulate_removeClause(s, (vecp_begin(&(s->clauses)))[i]);
		  change=0,noConflict=1;
		  if(solver_propagate(s)) {
			  solver_delete(s);
			  return true;
		  }

		  /* Update the clause w.r.t. the current assignment */
		  //if(!(refreshClause(s,&currentClause))){
		  if(!(refresh = refreshClause(s,&currentClause))){
			free_clause(currentClause);
			actualCurrentNumClause--;
			continue;
		  }
		  else if (refresh == -1) return false;
          		  
		  size=clause_size(currentClause);
		  num_assuming=0;
		  /* The MOMS-like sort of the clause has been removed */
		  /* MOMS_sort(currentClause); */
		  /* Two randomly selected literals are reversed */
		  p1=random()%(clause_size(currentClause));
		  p2=random()%(clause_size(currentClause));
		  reverseLit(currentClause,p1,p2);
		  /* Assignment of the opposite of each literal from the clause */
		  for(j=0;(j<size) && (noConflict) && (!change);j++){	  
			if(clause_size(currentClause)==1) {break;}
			lit=clause_giveLit(currentClause,j);
			num_assuming++;
			assume(s,negation(lit)); 
			noConflict=((learnt_cl=solver_propagate(s))==NULL);	
			/* The following line can be safely uncommented 
			   (provided that '}' is added just before the "if(!noConflict){" line). 
			   This would remove the next loop in case of conflict. 
			   However, performing this test appears to degrade performances. (weird...) */
			//if(noConflict){
			for(k=j+1;k<size;k++){
			  my_lit=clause_giveLit(currentClause,k);
			  // The literal is assigned to the same value. the remaining part of the clause can be "omitted" 
			  if(s->assigns[lit_var(my_lit)]!=lB_Undef){
				if(AmISat(my_lit)){
				  if(clause_size(currentClause)!=num_assuming+1){
				change=1; 
				clause_begin(currentClause)[num_assuming]=my_lit;
				currentClause->size_learnt=(num_assuming+1)<<1;
				  }else{
				// In this case, the clause is redundant but no shorter clause can be deduced 
				// (last literal already assigned) 
				j=size;
				break;
				  }
				}
				// The literal is assigned to the opposite value. it can be removed from the clause 
				else{ 
				  for(l=k+1;l<size;l++) 
				clause_begin(currentClause)[l-1]=clause_giveLit(currentClause,l); 
				  currentClause->size_learnt=(size-1)<<1; 
				  change=1;
				}
				break;
			  }
			}
		  }
		  if(!noConflict){
			if(clause_size(currentClause)!=num_assuming){
				change=1; 
				currentClause->size_learnt=(num_assuming)<<1;
			}
			veci_new(&learnt_clause);
			veci_resize(&learnt_clause,0);
			solver_analyze(s, learnt_cl, &learnt_clause);
			begin = veci_begin(&learnt_clause);
			end   = begin + veci_size(&learnt_clause);
			if(veci_size(&learnt_clause) > 1){
				#ifndef DONT_LEARN_ANYTHING
					  if(veci_size(&learnt_clause) < clause_size(currentClause)){
						c     = clause_new(s,begin,end,1);
						vecp_push(&s->learnts,c);
						s->stats.learnts++;
					  }
				#endif
			}
			else {
			  solver_canceluntil(s,s->root_level);
			  num_unit++;
			  enqueue(s,*begin,0);
			}
			if(clause_size(currentClause)!=num_assuming){
				currentClause->size_learnt=(num_assuming)<<1,change=1;
			}
		  }
		  solver_canceluntil(s,s->root_level);
		  if(clause_size(currentClause)==1){			
			num_unit++;
			enqueue(s,clause_giveLit(currentClause,0),(clause*)0); 
			free_clause(currentClause);
			actualCurrentNumClause--;
			continue; 
		  }
		  simulate_addItAgain(s,(vecp_begin(&(s->clauses)))[i],currentClause);
		  if(change){
			if(!timeOut) {globalModif=1;  lastModif=i+1;} 			
		  }
		}
	}    
  
	if(solver_propagate(s)) {
		solver_delete(s);
		return true;
	}

	solver_canceluntil(s,s->root_level);
  
	//fprintf(stderr,"#ITERATIONS %d\n",nbIteration);
	//fprintf(stderr,"#LEARNT CLAUSES %d\n",(int) s->stats.learnts);
  
	/* Export produced CNF */
	/* Detected unit clauses first, then shortened clauses from the original CNF, then learnt clauses (3 loops) */
	/*  
	printf("p cnf %d %d\n",s->size,(int) ((actualCurrentNumClause)+((int) s->stats.learnts)+countExportedUnit(s)));
	for(i=0,size=s->size;i<size;i++){
		if(s->assigns[i]!=lB_Undef){
		  printf("%d 0\n",s->assigns[i]==lB_True?i+1:-(i+1));
		}      
	}
	
	for(i=0,size=s->stats.clauses;i<size;i++){
		currentClause= (clause*) (vecp_begin(&(s->clauses)))[i];
		if(currentClause){
		  k=clause_size(currentClause);
		  for(j=0;j<k;j++) printf("%d ",wellWritten(clause_giveLit(currentClause,j)));
		  printf("0\n");
		}
	} 
	*/

	//Compute number of clauses and total length to determine	further simplification
	lUnitclauses = lUnitclauses_new; 
	lclauses = lclauses_new;
	llits = llits_new;
	lUnitclauses_new = 0;
	for(i=0,size=s->size;i<size;i++){
		if(s->assigns[i]!=lB_Undef){
		  lUnitclauses_new++;
		}      
	}

	lclauses_new = s->stats.clauses;
	llits_new = 0;
	for(i=0; i<lclauses_new; i++){
		currentClause= (clause*) (vecp_begin(&(s->clauses)))[i];
		if(currentClause){
		  llits_new += clause_size(currentClause);       
		}
	}    

  } /* My while */
  
  CNF_extract(cs, s, lUnitclauses_new, lclauses_new, llits_new);
  
  return true;
}

bool CNF_extract(CState* cs, solver* s, int Unitclauses_new, int clauses_new, int lits_new) {
	int i, size, k, j;
	int lUnitclauses, lclauses, llits=0;
	set<CSClause*>::iterator its;
	lUnitclauses = cs->m_L->m_L.count();
	lclauses = cs->m_clauses.size();
	for (its = cs->m_clauses.begin(); its != cs->m_clauses.end(); its++)
		llits += (*its)->m_literals.size();

	if ( Unitclauses_new == lUnitclauses && clauses_new == lclauses && lits_new == llits) {
		//cout << "No simplification" << endl;
		solver_delete(s);
		return false;
	}	
	cs->clear();
	CSClause* ns;
	clause *currentClause;
	//Literals U, NU;
	//Literals::iterator itl;
	//Literal l, nl;


	for(i=0,size=s->size;i<size;i++){
		if(s->assigns[i]!=lB_Undef){
		  //l = s->assigns[i]==lB_True? 2*i : 2*i+1;
		  //U.insert(l);
		  //NEGATE(l,nl);
		  //NU.insert(nl);
		  cs->m_L->m_L[s->assigns[i]==lB_True? 2*i : 2*i+1] = 1;
		}
	}	
	
	for(i=0,size=s->stats.clauses;i<size;i++){
		currentClause= (clause*) (vecp_begin(&(s->clauses)))[i];
		if(currentClause){
		  k=clause_size(currentClause);
		  ns = new CSClause;
		  for(j=0;j<k;j++) 
			  ns->m_literals.insert(cpaLit(clause_giveLit(currentClause,j)));		  
		  if (!cs->addClause(ns, true)) delete ns;
		}
	}

	//for(itl=U.begin(); itl!=U.end(); itl++) {
	//	if (cs->is_inClause(*itl)) cout<<"subsumed by Unit clause";
	//}
	
	//for(itl=NU.begin(); itl!=NU.end(); itl++) {
	//	if (cs->is_inClause(*itl)) cout<<"redundant by Unit clause";
	//}

	solver_delete(s);
	return true;
}

void CNF_extract(set<Literals>* cnf, solver* s , int Unitclauses_new, int clauses_new, int lits_new) {

	int i, size, k, j;	
	Literals st;		
	clause *currentClause;
	
	int lUnitclauses=0, lclauses=0, llits=0;
	set<Literals>::iterator its;

	for (its = cnf->begin(); its != cnf->end(); its++) {
		if (its->size() == 1)
			lUnitclauses += 1;
		else {
			lclauses += 1;
			llits += its->size();
		}
	}

	if ( Unitclauses_new == lUnitclauses && clauses_new == lclauses && lits_new == llits) {
		//cout << "No simplification" << endl;
		solver_delete(s);
		return;
	}
	
	cnf->clear();

	for(i=0,size=s->size;i<size;i++){
		if(s->assigns[i]!=lB_Undef){
		  st.clear();	
		  st.insert(s->assigns[i]==lB_True? 2*i : 2*i+1);
		  cnf->insert(st);
		}
	}

	set<Literals> ss;	

	bool subsumed;
	
	for(i=0,size=s->stats.clauses;i<size;i++){
		currentClause= (clause*) (vecp_begin(&(s->clauses)))[i];
		if(currentClause){
		  k=clause_size(currentClause);
		  st.clear();
		  for(j=0;j<k;j++) 
			  st.insert(cpaLit(clause_giveLit(currentClause,j)));
		  if (cnf->count(st))
			  continue;
		  ss.clear();

		  subsumed = false;
		  
		  for (its = cnf->begin(); its != cnf->end(); its++) {
			  if (subsumes(*its,st)) {
				  subsumed = true;
				  break;
			  }
			  if (subsumes(st, *its))
				  ss.insert(*its);
		  }

		  if (!subsumed) {
			  for (its = ss.begin(); its != ss.end(); its++)
				  cnf->erase(*its);
			  cnf->insert(st);
		  }
		}
	}
	solver_delete(s);
}


bool subsumes(const Literals& s, const Literals& s1) {

	if (s.size() >= s1.size())
		return false;	

	Literals::const_iterator it, it1;

	for (it = s.begin(), it1 = s1.begin(); it != s.end(); it++) {
		for ( ; it1 != s1.end(); it1++) {
			if (*it1 >= *it)
				break;
		}
		if (it1 == s1.end())
			return false;
		if (*it1 != *it)
			return false;
	}
	return true;
}


void getNegationCNF(set<Literals>* cnf_in, set<Literals>* cnf_out) {
	set<Literals>::iterator its1, its2;
	set<Literals> cnf_temp;
	Literals::iterator itl;
	Literals st;
	Literal l;
	
	its1 = cnf_in->begin();
	for(itl = its1->begin(); itl != its1->end(); itl++) {
		NEGATE(*itl, l);
		st.insert(l);
		cnf_out->insert(st);
		st.clear();
	}

	for (its1++; its1 != cnf_in->end(); its1++) {
		cnf_temp = *cnf_out;
		cnf_out->clear();
		for (its2 = cnf_temp.begin(); its2 != cnf_temp.end(); its2++) {
			for(itl = its1->begin(); itl != its1->end(); itl++) {
				if (its2->find(*itl) == its2->end()) {
					NEGATE(*itl, l);
					st = *its2;
					st.insert(l);
					cnf_out->insert(st);
					st.clear();
				}
			}
		}
	}
}

//=========================================================================
static inline void s_removeClause(solver *s, void *c){
  lit* lits = clause_begin((clause*) c);
  vecp_remove(solver_read_wlist(s,lit_neg(lits[0])),(void*)(clause_size((clause*) c) > 2 ? c : clause_from_lit(lits[1])));
  vecp_remove(solver_read_wlist(s,lit_neg(lits[1])),(void*)(clause_size((clause*)c) > 2 ? c : clause_from_lit(lits[0])));
}

static inline void s_addItAgain(solver *s, clause *c){
  vecp_push(solver_read_wlist(s,lit_neg(c->lits[0])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(c->lits[1])));
  vecp_push(solver_read_wlist(s,lit_neg(c->lits[1])),(void*)(clause_size(c) > 2 ? c : clause_from_lit(c->lits[0])));
}


