#include "cnfstate.h"
#include "planner.h"
#include "parsingFunctions.h"

/*************************************************************************
 * State Implementation
 *************************************************************************/
State::State(const Planner* planner)
{
	m_L.resize(planner->m_literals.size());
	m_planner = planner;
	m_gchecked = -1;
	m_hvalue = -1;
	m_hash = -1;
	m_allocated = false;
}
  
State::~State()
{
  
}

void State::clear()
{  
  m_L.reset();
  m_gchecked = -1;
  m_hvalue = -1;
  m_hash = -1;	
}

/* operators */
int State::operator<(const State& s) const {
  return m_L < s.m_L;
}

bool State::operator==(const State& s) const {
  return m_L == s.m_L;
}

State State::operator=(const State& s) {
  m_L = s.m_L;
  m_planner = s.m_planner;
  return *this;
}


/* other functions */
bool State::is_consistent() const
{
	int i;

	for (i = 0; i < m_L.size()-1; i += 2) {
		if (m_L[i] && m_L[i+1]) return false;
	}
	return true; 
}

bool State::satisfies(const Literals* x)
{
  Literals::const_iterator it;  
  for (it = x->begin(); it != x->end(); it++) {	
	if (m_L[*it]) return true;	
  }
  return false;
}


bool State::goal_satisfied()
{
  if (m_gchecked != -1)
	  return m_gchecked;  

  set<Literals>::const_iterator it;
  int i;
  
  m_gchecked = 1;
 
  m_gsatisfied.resize(m_planner->m_gd.size(),1);

  // goal satisfaction has not been checked
  for (it = (m_planner->m_gd).begin(), i=0;
       it != (m_planner->m_gd).end(); it++, i++) {    
    if (!satisfies(&(*it))) {	  
      m_gsatisfied[i] = false;
	  m_gchecked = 0;
    }
  } 

  return m_gchecked;
}

bool State::intersect(const Literals* x) const
{
  Literals::const_iterator it_x;
  
  for (it_x = x->begin(); it_x != x->end(); it_x++) {
	if (m_L[*it_x])
	  return true;
  }
  
  return false;
}

bool State::intersect(const State* s) const
{
  return (m_L & s->m_L).count();
}

bool State::includes(const Literals* x) const
{
  Literals::const_iterator it_x;  
  for (it_x = x->begin(); it_x != x->end(); it_x++) {
	  if (!m_L[*it_x])
	  return false;
  }

  return true;
}

bool State::includes(const State* s) const
{
  return s->m_L.is_subset_of(m_L);
}


int State::hgvalue() 
{
  if (m_hvalue == -1) {
    Literals::iterator it;
	m_hvalue = 0;	
	
	for (it = m_planner->m_goal.begin(); it != m_planner->m_goal.end(); it++) {
		if (m_L[*it])
			m_hvalue++;
	}
  }  

  return m_hvalue;
}

int State::hvalue() {
	if (m_hvalue == -1) {  
		m_hvalue = m_gsatisfied.count();
	}
	return m_hvalue;
}

/*
int State::hvalue()
{
  if (m_hvalue == -1) {    	
	set<Literals>::iterator its;
	m_hvalue = 0;	

	for (its = m_planner->m_gd.begin(); its != m_planner->m_gd.end(); its++)
		if (this->intersect(&(*its)))
			m_hvalue++;	  
  }

  return m_hvalue;
}
*/

int State::hash()
{
  if (m_hash == -1) {

    Literals::iterator it;
	unsigned int l;
	unsigned int hash = 0;
	
	l = m_L.find_first();
	while (l < m_L.size()) {
		hash += (++l)*l;
		l = m_L.find_next(l);
	}

	m_hash = hash & HASH_BITS;
  }  

  return m_hash;
}


void State::print() const
{
	Literals L;
	unsigned int l;
	l = m_L.find_first();
	while (l < m_L.size()) {
		L.insert(l);
		l = m_L.find_next(l);
	}

	cout << "{";
		m_planner->print(L);  
	cout << "}";
}

/*************************************************************************
 * Clause Implementation
 *************************************************************************/
CSClause::CSClause()
{	
	m_hash = -1;	
	m_allocated = false;
}
  
CSClause::~CSClause()
{
  
}

void CSClause::clear()
{  
  m_literals.clear();  
  m_hash = -1;	
}

/* operators */
int CSClause::operator<(const CSClause& s) const {
  return m_literals < s.m_literals;
}

bool CSClause::operator==(const CSClause& s) const {
  return m_literals == s.m_literals;
}

CSClause CSClause::operator=(const CSClause& s) {
  m_literals = s.m_literals;  
  return *this;
}


/* other functions */


bool CSClause::is_consistent() const {
	Literals::const_iterator it_x, it_x1;
	Literal l;

	for (it_x = m_literals.begin(),it_x1 = it_x,it_x1++; it_x1 != m_literals.end(); it_x++,it_x1++) {
		NEGATE(*it_x, l);
		if (l == *it_x1) return false;
	}
	return true;
}

bool CSClause::intersect(const Literals* x) const
{
  Literals::const_iterator it_x;
  
  if (m_literals.size() > x->size()) {
	  for (it_x = x->begin(); it_x != x->end(); it_x++) {
		if (m_literals.count((*it_x)))
		  return true;
	  }
  }
  else {
	  for (it_x = m_literals.begin(); it_x != m_literals.end(); it_x++) {
		if (x->count((*it_x)))
		  return true;
	  }
  }

  return false;
}

bool CSClause::intersect(const CSClause* s) const
{
  return intersect(&s->m_literals);
}

bool CSClause::includes(const Literals* x) const
{
  return std::includes(m_literals.begin(),m_literals.end(),
					   x->begin(),x->end());
}

bool CSClause::includes(const CSClause* s) const
{
  return this->includes(&s->m_literals);
}

bool CSClause::subsumes(const CSClause* s) const
{
	if (m_literals.size() >= s->m_literals.size())
		return false;	

	Literals::const_iterator it, it1;

	for (it = m_literals.begin(), it1 = s->m_literals.begin(); it != m_literals.end(); it++) {
		for ( ; it1 != s->m_literals.end(); it1++) {
			if (*it1 >= *it)
				break;
		}
		if (it1 == s->m_literals.end())
			return false;
		if (*it1 != *it)
			return false;
	}
	return true;
}

bool CSClause::subsumes(const Literals* s) const
{		
	if (!m_literals.size() || m_literals.size() > s->size())
		return false;	
	
	Literals::const_iterator it, it1;	
	for (it = m_literals.begin(), it1 = s->begin(); it != m_literals.end(); it++) {		
		for ( ; it1 != s->end(); it1++) {
			if (*it1 >= *it)
				break;
		}		
		if (it1 == s->end())
			return false;		
		if (*it1 != *it)
			return false;		
	}
	
	return true;
}

int CSClause::hash()
{
  if (m_hash == -1) {

    Literals::iterator it;
	unsigned int l;
	unsigned int hash = 0;
	
	for (it = m_literals.begin(); it != m_literals.end(); it++) {		
		l = *it;
		hash += (++l)*l;
	}
	m_hash = hash & HASH_BITS;
  }  

  return m_hash;
}


void CSClause::print() const
{
  cout << "{";
	//m_planner->print(m_literals);
  //cout<<this;
  cout << "}";
}


/*************************************************************************
 * CState Implementation
 *************************************************************************/
/* constructor and destructor */
//CState::CState(const Planner* planner)
CState::CState()
{
  //m_L = new State(m_planner);
  //m_planner = planner;  
  m_plan_length = 0;
  m_action = NULL;
  m_previous_cstate = NULL;
  m_hvalue = -1;
  m_size = -1;
  m_hash = -1;
  m_allocated = false;
  
  m_status = UNEXPANDED;  
  m_active_in_transitions = 1;
  m_depth = 0;
}

CState::~CState() 
{
	
}

void CState::clear()
{   
  m_L->clear();    
  set<CSClause*>::iterator its;
  for (its = m_clauses.begin(); its != m_clauses.end(); its++) {
	  if (!(*its)->m_allocated)
		delete (*its);	
  }
  m_clauses.clear();  
  m_inS.clear();
  m_hvalue = -1;
  m_size = -1;
  m_hash = -1;
}

bool CState::operator<(const CState& cs) const 
{
  if (m_L == cs.m_L)
	return m_clauses < cs.m_clauses;
  else
    return m_L < cs.m_L;
}

/*
CState CState::operator=(const CState& cs)
{ 
  m_L = new State(m_planner);
  m_L->m_L = cs.m_L->m_L;
  m_EF = cs.m_EF;    

  set<CSClause*>::const_iterator it_s;
  CSClause* s;
  for (it_s = cs.m_clauses.begin(); it_s != cs.m_clauses.end(); it_s++) { 
	  s = new CSClause;
	  s->m_literals = (*it_s)->m_literals;
	  m_clauses.insert(s);
  }    

  this->inClauses();  
  return *this;
}
*/

/*
void CState::copy(const CState* cs)
{ 
  m_planner	= cs->m_planner;
  m_L = new State(m_planner);
  m_L->m_L = cs->m_L->m_L;
  m_EF = cs->m_EF;
  
  m_clauses = cs->m_clauses;
  m_inS = cs->m_inS;

}
*/

void CState::copy(const CState* cs)
{ 
  m_planner	= cs->m_planner;
  m_L = new State(m_planner);
  m_L->m_L = cs->m_L->m_L;
  m_EF = cs->m_EF;


  set<CSClause*>::const_iterator it_s;
  CSClause* s;
  for (it_s = cs->m_clauses.begin(); it_s != cs->m_clauses.end(); it_s++) { 
	  s = new CSClause;
	  s->m_literals = (*it_s)->m_literals;
	  m_clauses.insert(s);
  }    

  this->inClauses();  
}


void CState::preprocess()
{
	if (!m_clauses.size()) return;
	set<CSClause*> l_clauses, P, P1;
	vector<CSClause*> R, U, unchecked;
	set<CSClause*>::iterator its, its1;
	vector<CSClause*>::iterator itvs;
	CSClause *s, *sr;
	Literals::iterator itl;
	Literal l;
	int i;
	bool changed = true, changed1, restart;
	
	int t = m_planner->THRESHOLD;
	m_planner->THRESHOLD = 0;
	inClauses();
	m_planner->THRESHOLD = t;
		
	for ( l = m_L->m_L.find_first(); l < m_L->m_L.size(); ) {
		s = new CSClause;
		s->m_literals.insert(l);
		m_clauses.insert(s);
		inClause(s);
		U.push_back(s);		
		l = m_L->m_L.find_next(l);
	}	
	
	while (changed) {
		changed = false;
		//sort the clauses and store in unchecked
		l_clauses = m_clauses;		
		while (l_clauses.size()) {
			for (its = l_clauses.begin(), its1=its++; its != l_clauses.end(); its++) {
				if ((*its1)->m_literals.size() < (*its)->m_literals.size()) its1 = its;				
			}
			unchecked.push_back(*its1);
			l_clauses.erase(its1);
		}		
		while (unchecked.size()) {					
			s = *unchecked.rbegin();
			unchecked.pop_back();			
			for (itl = s->m_literals.begin(); itl!=s->m_literals.end(); itl++) {
				NEGATE(*itl, l);
				P = m_inS[l];
				P1 = P;				
				for (its = P.begin(); its != P.end(); its++) {										
					if (!P1.count(*its)) continue;
					restart = false;
					//cout << "8 " ; fflush(stdout);
					if ((sr = resolvent(s, *its, *itl, l)) && checkSubsumption(sr, &P1, &unchecked)) {						
						//cout << "9 " ; fflush(stdout);
						changed = true;
						m_clauses.insert(sr); 
						inClause(sr);
						if (sr->m_literals.size() == 1)	
							U.push_back(sr);
						if (!m_clauses.count(s)) {
							unchecked.push_back(sr);
							restart = true;
							break;
						}
						else if (unchecked.size()) {
							for (itvs = unchecked.begin(); itvs != unchecked.end(); itvs++)
								if ((*itvs)->m_literals.size() <= sr->m_literals.size()) break;
							if (itvs != unchecked.end()) 
								unchecked.insert(itvs,sr);
							else
								unchecked.push_back(sr);
						}
						else
							unchecked.push_back(sr);
					}
				}
				if (restart) break;
			}
		}		
	}	
	//Update unit clauses
	for (itvs = U.begin(); itvs != U.end(); itvs++) {
		m_L->m_L[*(*itvs)->m_literals.begin()]=1;
		removeClause(*itvs);
		if (!(*itvs)->m_allocated)
			delete *itvs;
	}	
	if (!m_clauses.size())
		m_inS.clear();		
}

CSClause* CState::resolvent(const CSClause* s, const CSClause* s1, const Literal l, const Literal l1) {	
	Literals::iterator itl;
	Literal l2;
	
	if (s->m_literals.size() <= s1->m_literals.size()) {		
		//Check for tautology
		for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {
			if (*itl != l) {
				NEGATE(*itl, l2);
				if (s1->m_literals.count(l2))
					return NULL;
			}
		}		
		CSClause* s_new;
		s_new = new CSClause;
		s_new->m_literals = s1->m_literals;
		s_new->m_literals.erase(l1);
		s_new->m_literals.insert(s->m_literals.begin(), s->m_literals.end());
		s_new->m_literals.erase(l);		
		return s_new;
	}
	else {
		//Check for tautology
		for (itl = s1->m_literals.begin(); itl != s1->m_literals.end(); itl++) {
			if (*itl != l1) {
				NEGATE(*itl, l2);
				if (s->m_literals.count(l2))
					return NULL;
			}
		}
		CSClause* s_new; 
		s_new = new CSClause;
		s_new->m_literals = s->m_literals;
		s_new->m_literals.erase(l);
		s_new->m_literals.insert(s1->m_literals.begin(), s1->m_literals.end());
		s_new->m_literals.erase(l1);		
		return s_new;
	}
}


bool CState::checkSubsumption(const CSClause* s, set<CSClause*>* P, vector<CSClause*>* unchecked) {
	set<CSClause*>::iterator its;
	vector<CSClause*>::iterator itvs;
	
	Literals::iterator itl;
	int i;
	vector<CSClause*> subsumed_set;
		
	if (m_inS.size()) {
		Literal l;
		int min = -1;	

		for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {
			if (min > m_inS[*itl].size() || min == -1) {
				min = m_inS[*itl].size();
				l = *itl;
				if (min == 0)
					break;
			}
		}
		for (its = m_inS[l].begin(); its != m_inS[l].end(); its++) {								
			if (s->subsumes(*its))
				subsumed_set.push_back(*its);							
		}		
	}
	else {		
		for (its = m_clauses.begin(); its != m_clauses.end(); its++) {			
			if (s->subsumes(*its)) 				
				subsumed_set.push_back(*its);										
		}		
	}
	
	if (subsumed_set.size()) {
		for (i = 0; i<subsumed_set.size(); i++) {
			removeClause(subsumed_set[i]);
			outClause(subsumed_set[i]);
			P->erase(subsumed_set[i]);
			for (itvs = unchecked->begin(); itvs != unchecked->end(); itvs++) {
				if (*itvs == subsumed_set[i]) {
					unchecked->erase(itvs);
					break;
				}
			}
			if (!subsumed_set[i]->m_allocated)
				delete subsumed_set[i];
		}		
		return true;
	}	
	return false;
}

bool CState::removeSubsumed(const CSClause* s) {
	set<CSClause*>::iterator its;
	vector<CSClause*>::iterator itvs;
	
	Literals::iterator itl;
	int i;
	vector<CSClause*> subsumed_set;
		
	if (m_inS.size()) {
		Literal l;
		int min = -1;	

		for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {
			if (min > m_inS[*itl].size() || min == -1) {
				min = m_inS[*itl].size();
				l = *itl;
				if (min == 0)
					break;
			}
		}
		for (its = m_inS[l].begin(); its != m_inS[l].end(); its++) {								
			if (s->subsumes(*its))
				subsumed_set.push_back(*its);							
		}		
	}
	else {		
		for (its = m_clauses.begin(); its != m_clauses.end(); its++) {			
			if (s->subsumes(*its)) 				
				subsumed_set.push_back(*its);										
		}		
	}
	
	if (subsumed_set.size()) {
		for (i = 0; i<subsumed_set.size(); i++) {
			removeClause(subsumed_set[i]);
			outClause(subsumed_set[i]);
			if (!subsumed_set[i]->m_allocated) {				
				delete subsumed_set[i];
			}
		}		
		return true;
	}	
	return false;
}



void CState::removeClause(CSClause* s) {
	Literals::iterator itl;	
	
	if (m_clauses.erase(s) && m_inS.size()) {	
		for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {
			//cout << ":" <<*itl<<"<"<<m_inS.size()<<" "; fflush(stdout);
			m_inS[*itl].erase(s);
		}
	}
}

void CState::inClause(CSClause* s) {
	Literals::iterator itl;
		
	//m_clauses.erase(s);
	if (m_inS.size()) {
		for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {			
			m_inS[*itl].insert(s);
		}
	}
}

void CState::outClause(CSClause* s) {
	Literals::iterator itl;
	
	//m_clauses.erase(s);
	if (m_inS.size()) {
		for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {			
			m_inS[*itl].erase(s);
		}
	}
}


bool CState::addClause(CSClause* s, const bool checkSubsumed) {
	set<CSClause*>::iterator its;
	Literals::iterator itl;
	int i;

	/*
	if (!s->is_consistent()) {
		cout <<endl << "inconsistent clause!" <<endl; fflush(stdout);
		//exit(1);
	}

	if (s->m_literals.size() < 2) {
		cout <<endl << "WRONG clause size = !"<< s->m_literals.size() <<endl; fflush(stdout);
		exit(1);
	}
	*/

	vector<CSClause*> subsumed_set;

	if (m_inS.size()) {
		Literal l;
		int min = -1;	

		//cout << "3"; fflush(stdout);

		for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {
			if (min > m_inS[*itl].size() || min == -1) {
				min = m_inS[*itl].size();
				l = *itl;
				if (min == 0)
					break;
			}
		}
		
		//cout << "4"; fflush(stdout);

		for (its = m_inS[l].begin(); its != m_inS[l].end(); its++) {		
			if (checkSubsumed) {
				if (s->m_literals == (*its)->m_literals){
					//If the state already exists so do nothing
					//delete s;					
					return false;
				}
			}
			
			if (s->subsumes(*its)) {						
				subsumed_set.push_back(*its);
			}
		}
		//cout << "5"; fflush(stdout);
	}
	else {
		//cout << "6"; fflush(stdout);
		for (its = m_clauses.begin(); its != m_clauses.end(); its++) {
			if (checkSubsumed) {
				if (s->m_literals == (*its)->m_literals){
					//If the state already exists so do nothing
					//delete s;					
					return false;
				}
			}
			
			if (s->subsumes(*its)) {						
				subsumed_set.push_back(*its);
			}
		}
		//cout << "7"; fflush(stdout);
	}


	if (subsumed_set.size()) {
		//cout << "8"; fflush(stdout);
		for (i = 0; i<subsumed_set.size(); i++) {
			removeClause(subsumed_set[i]);
			if (!subsumed_set[i]->m_allocated)
				delete subsumed_set[i];			
		}

		//cout << "9"; fflush(stdout);
		m_clauses.insert(s);
		if (m_inS.size()) {
			for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++)
				m_inS[*itl].insert(s);
		}
		//cout << "a"; fflush(stdout);
		return true;
	}
	else if (checkSubsumed && s->m_literals.size()>2) {
		//cout << "b"; fflush(stdout);
		if (this->subsumes(s)) {
			//delete s;			
			return false;
		}		
	}
		
	m_clauses.insert(s);
	if (m_inS.size()) {		
		inClause(s);
	}
	else {		
		inClauses();		
	}	

	return true;
}
//NEW TEST

void CState::allocate() {
	if (m_allocated) {cout <<"allocated!"; return;}
	vector<CSClause*> vs;
	vector<CSClause*>::iterator its;
	Literals::iterator itl;
	CSClause* s;
	int i;
	bool update = false;
	
	m_L = m_planner->add_state(m_L);
	vs.assign(m_clauses.begin(), m_clauses.end());		
	m_clauses.clear();		
	for(its = vs.begin(); its != vs.end(); its++) {	
		s = m_planner->add_clause(*its);				
		m_clauses.insert(s);
		
		if (s != *its && !update) update = true;

		/*
		if (s != *its && m_inS.size()) {
			for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {
				m_inS[*itl].erase(*its);
				m_inS[*itl].insert(s);
			}
		}*/

	}
	
	//if ((update || (m_clauses.size() < 2*m_planner->THRESHOLD)) && m_inS.size()) m_inS.clear();
	if (m_inS.size()) m_inS.clear();
	//inClauses();
	m_allocated = true;
}


/*
void CState::allocate() {
	if (m_allocated) {cout <<"allocated!"; return;}
	set<CSClause*> diff_ad, diff_rm;
	//set<CSClause*>::iterator its;
	vector<CSClause*> vs;
	vector<CSClause*>::iterator its;
	Literals::iterator itl;
	CSClause* s;
	int i;

	m_L = m_planner->add_state(m_L);
	vs.assign(m_clauses.begin(), m_clauses.end());
	for(its = vs.begin(); its != vs.end(); its++) {
		s = m_planner->add_clause(*its);
		if (s != *its) {
			m_clauses.erase(*its);
			m_clauses.insert(s);
			if (m_inS.size()) {
				for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {
					m_inS[*itl].erase(*its);
					m_inS[*itl].insert(s);
				}

			}			
		}
	}	
	
	m_allocated = true;
}
*/

//Build m_inS structure once. Increments should be maintain at different places
void CState::inClauses() {	
	set<CSClause*>::const_iterator it_s;
	Literals::const_iterator it_l;
	int i;

	if (m_inS.size() || m_clauses.size() < m_planner->THRESHOLD) return;

	m_inS.resize(m_planner->m_literals.size());
	
	for (it_s = this->m_clauses.begin(); it_s != this->m_clauses.end(); it_s++) {
		for (it_l = (*it_s)->m_literals.begin(); it_l != (*it_s)->m_literals.end(); it_l++) {
			m_inS[*it_l].insert(*it_s);	
		}
	}
}

int CState::Unit_Extract() {	
	Literals::iterator itl;	
	int i=0;

	for(itl = m_planner->m_literals.begin(); itl != m_planner->m_literals.end(); itl++) {
		if (!m_L->m_L[*itl] && entails(*itl)) {
			m_L->m_L[*itl] = 1;
			UP(*itl);
			i++;
		}
	}
	if (i)
	cout << "Found " << i << "new unit literal(s)";
	return i;
}


void CState::UP() {
	list<Literal> U;
	Literals::iterator itl;
	Literal l, l1;
	CSClause* s, *s1;		

	for(itl = m_planner->m_literals.begin(); itl != m_planner->m_literals.end(); itl++) {
			if (m_L->m_L[*itl]) 
				U.push_back(*itl);
	}

	while (U.size()) {		
		//remove claused subsumed by the unit-clause
		l = *(U.begin());
		U.pop_front();
		while (s = is_inClause(l)) {
			removeClause(s);									
			if (!s->m_allocated) {
				//cout << "Test 1 "; fflush(stdout);
				delete s;
			}
			//cout << " 2 "; fflush(stdout);
		}
		NEGATE(l, l1);
		while (s = is_inClause(l1)) {
			if (s->m_allocated) {
				s1 = new CSClause;
				s1->m_literals = s->m_literals;				
				removeClause(s);				
				s = s1;
				inClause(s);
			}

			if (s->m_literals.size() > 2) {
				s->m_literals.erase(l1);
				if (m_inS.size())
					m_inS[l1].erase(s);
				removeSubsumed(s);				
			}
			else {
				if (*s->m_literals.begin() == l1) {					
					m_L->m_L[*s->m_literals.rbegin()] = 1;
					U.push_back(*s->m_literals.rbegin());
				}
				else {
					m_L->m_L[*s->m_literals.begin()] = 1;								
					U.push_back(*s->m_literals.begin());
				}								
				removeClause(s);							
				if (!s->m_allocated) {
					//cout << "Test 3 "; fflush(stdout);
					delete s;
				}
				//cout << " 4 "; fflush(stdout);
			}
		}		
	}	
}

void CState::UP(const Literal pl) {
	list<Literal> U;
	Literals::iterator itl;
	Literal l, l1;
	CSClause* s, *s1;	

	U.push_back(pl);
	if (!m_L->m_L[pl])
		m_L->m_L[pl] = 1;

	while (U.size()) {		
		//remove claused subsumed by the unit-clause
		l = *(U.begin());
		U.pop_front();
		while (s = is_inClause(l)) {
			removeClause(s);						
			if (!s->m_allocated)
				delete s;
		}
		NEGATE(l, l1);
		while (s = is_inClause(l1)) {
			if (s->m_allocated) {
				s1 = new CSClause;
				s1->m_literals = s->m_literals;
				removeClause(s);
				s = s1;
				inClause(s);
			}

			if (s->m_literals.size() > 2) {
				s->m_literals.erase(l1);
				if (m_inS.size())
					m_inS[l1].erase(s);
				removeSubsumed(s);
			}
			else {
				if (*s->m_literals.begin() == l1) {					
					m_L->m_L[*s->m_literals.rbegin()] = 1;
					U.push_back(*s->m_literals.rbegin());
				}
				else {
					m_L->m_L[*s->m_literals.begin()] = 1;								
					U.push_back(*s->m_literals.begin());
				}								
				removeClause(s);			
				if (!s->m_allocated)
					delete s;
			}
		}		
	}	
}

void CState::UP(const Literals& st) {
	list<Literal> U;
	Literals::iterator itl;
	Literal l, l1;
	CSClause* s, *s1;	

	U.assign(st.begin(), st.end());
	for (itl = st.begin(); itl != st.end(); itl++) {
		m_L->m_L[*itl] = 1;
	}
	
	//cout << "Test 0 "; fflush(stdout);

	while (U.size()) {		
		//remove claused subsumed by the unit-clause
		l = *(U.begin());
		U.pop_front();
		NEGATE(l, l1);
		while (s = is_inClause(l)) {			
			removeClause(s);						
			if (!s->m_allocated) {					
				delete s;
			}			
		}
		
		while (s = is_inClause(l1)) {
			if (s->m_allocated) {
				s1 = new CSClause;
				s1->m_literals = s->m_literals;
				removeClause(s);
				s = s1;
				m_clauses.insert(s);				
				inClause(s);
			}

			if (s->m_literals.size() > 2) {
				s->m_literals.erase(l1);
				if (m_inS.size())
					m_inS[l1].erase(s);				
				removeSubsumed(s);				
			}
			else {
				if (*s->m_literals.begin() == l1) {					
					m_L->m_L[*s->m_literals.rbegin()] = 1;
					U.push_back(*s->m_literals.rbegin());
				}
				else {
					m_L->m_L[*s->m_literals.begin()] = 1;								
					U.push_back(*s->m_literals.begin());
				}								
				
				if (!m_clauses.count(s))
					cout <<" NOT in the cstate ";

				removeClause(s);			
				if (!s->m_allocated) {					
					//cout << " del " << s <<" "; fflush(stdout);
					delete s;
				}				
			}
		}		
	}		
}


void CState::and_merge(const Literal l) {		
	if (!m_L->m_L[l]) {		
		//m_L->m_L[l] = 1;
		//m_planner->m_timer->start(SIMPLIFICATION_TIMER);
		this->UP(l);
		simplifyCNF(this);		
		//m_planner->m_timer->end(SIMPLIFICATION_TIMER);
	}	
}

void CState::and_merge(const Literals* ls) {
	Literals::const_iterator it_l;	
	bool new_lit = false;	
	//m_planner->m_timer->start(SIMPLIFICATION_TIMER);
	for (it_l = ls->begin(); it_l!=ls->end();it_l++) {
		if (!m_L->m_L[*it_l]) {
			this->UP(*it_l);
			//m_L->m_L[*it_l] = 1;			
			if (!new_lit) new_lit = true;
		}
	}

	if (new_lit)
		simplifyCNF(this);
	//m_planner->m_timer->end(SIMPLIFICATION_TIMER);
}

//cs_new = this OR cs
void CState::or_merge(CState* cs,CState* cs_new ) {	
	Literals L;	
	set<CSClause*>::const_iterator it_s, it_s1;
	Literals::const_iterator it_l;
	set<CSClause*> S;
	CSClause* s;
	Literal l;
	
	//cout << "1"; fflush(stdout);	
	cs_new->m_L->m_L =  m_L->m_L & cs->m_L->m_L;
	
	m_L->m_L -= cs_new->m_L->m_L;
	cs->m_L->m_L -= cs_new->m_L->m_L;
	//cout << "2"; fflush(stdout);
	for ( l = cs->m_L->m_L.find_first(); l < cs->m_L->m_L.size(); ) {
		while(s = is_inClause(l)) {					
			if (cs_new->addClause(s, true)) 
				this->removeClause(s);
			else {
				this->removeClause(s);
				if (!s->m_allocated)
					delete s;
			}
		}
		l = cs->m_L->m_L.find_next(l);
	}	
	
	//cout << "3"; fflush(stdout);

	for ( l = m_L->m_L.find_first(); l < m_L->m_L.size(); ) {
		while(s = cs->is_inClause(l)) {					
			if (cs_new->addClause(s, true))
				cs->removeClause(s);
			else {
				cs->removeClause(s);
				if (!s->m_allocated)
					delete s;
			}

		}
		l = m_L->m_L.find_next(l);
	}
	
	//cout << "4"; fflush(stdout);
	for (it_s = cs->m_clauses.begin(); it_s != cs->m_clauses.end(); it_s++) {
		if (s = this->includeClause(*it_s)) {
			//s->print(); cout<<" = "; (*it_s)->print(); fflush(stdout);			
			if (cs_new->addClause(*it_s, true))
				cs->outClause(*it_s);
			else {
				cs->outClause(*it_s);
				if (!(*it_s)->m_allocated)
					delete *it_s;
			}
			S.insert(*it_s);
			this->removeClause(s);
			if (!s->m_allocated)
				delete s;			
		}
		
		else if (s = this->subsumes(*it_s)) {			
			if (cs_new->addClause(*it_s, true))
				cs->outClause(*it_s);
			else {
				cs->outClause(*it_s);
				if (!(*it_s)->m_allocated)
					delete *it_s;
			}
			S.insert(*it_s);
		}
		
	}

	//cout << "5"; fflush(stdout);
    for (it_s = S.begin(); it_s != S.end(); it_s++)
		cs->m_clauses.erase(*it_s);
		//cs->m_clauses.erase(S.begin(), S.end());
	S.clear();
	
	//cout << "6"; fflush(stdout);

	for (it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {
		if (s = cs->includeClause(*it_s)) {			
			if (cs_new->addClause(*it_s, true))
				this->outClause(*it_s);
			else {
				this->outClause(*it_s);
				if (!(*it_s)->m_allocated)
					delete *it_s;
			}				
			S.insert(*it_s);			
			cs->removeClause(s);
			if (!s->m_allocated)
				delete s;			
		}
		
		else if (s = cs->subsumes(*it_s)) {			
			if (cs_new->addClause(*it_s, true))
				this->outClause(*it_s);
			else {
				this->outClause(*it_s);
				if (!(*it_s)->m_allocated)
					delete *it_s;
			}				
			S.insert(*it_s);
		}
		
	}
	
	//cout << "7"; fflush(stdout);
	
	for (it_s = S.begin(); it_s != S.end(); it_s++)
		this->m_clauses.erase(*it_s);
	//this->m_clauses.erase(S.begin(), S.end());
	
	//cout << "8"; fflush(stdout);

	for (l = cs->m_L->m_L.find_first(); l < cs->m_L->m_L.size(); ) {
		//cout << "1"; fflush(stdout);
		this->or_merge(l,cs_new);
		//cout << "2"; fflush(stdout);
		l = cs->m_L->m_L.find_next(l);
	}	    		
	//cout << "3"; fflush(stdout);

	for (it_s = cs->m_clauses.begin(); it_s != cs->m_clauses.end(); it_s++) {		
		this->or_merge(*it_s,cs_new);		
	}	

	//cout << "9"; fflush(stdout);
				
	return;	
}

void CState::or_merge(Literal l, CState* cs_new) const {
	CSClause* s;		
	set<CSClause*>::const_iterator it_s;		
	Literal l1, li;
	NEGATE(l, l1);	
	
	//cout << "0"; fflush(stdout);

	for (li = m_L->m_L.find_first(); li < m_L->m_L.size(); ) {			
		if (li != l1) {
			s = new CSClause;
			s->m_literals.insert(li);
			s->m_literals.insert(l);			
			if (!cs_new->addClause(s, true)) 				
				delete s; //CHECK
			//cout << "2"; fflush(stdout);
		}		
		li = m_L->m_L.find_next(li);
	}	

	//cout << "3"; fflush(stdout);

	for (it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {		
		//if (!m_inS[l].count(*it_s)) {
		if (!(*it_s)->m_literals.count(l)) {
			//if (!m_inS[l1].count(*it_s)) {
			  if (!(*it_s)->m_literals.count(l1)) {
				s = new CSClause;
				s->m_literals = (*it_s)->m_literals;
				s->m_literals.insert(l);				
				if (!cs_new->addClause(s, true)) delete s;
			}
		}
		else {			
			s = new CSClause;
			s->m_literals = (*it_s)->m_literals;			
			if (!cs_new->addClause(s, true)) delete s;
		}		
	}
	//cout << "4"; fflush(stdout);
}


void CState::or_merge(CSClause* s1, CState* cs_new) const {
	
	CSClause* s;
	Literals::const_iterator it_l;
	set<CSClause*>::const_iterator it_s, it_s1;	
	Literal l, li;		
		
	for (li = m_L->m_L.find_first(); li < m_L->m_L.size(); ){		
		NEGATE(li, l);		
		if (s1->m_literals.count(l)) {
			li = m_L->m_L.find_next(li);
			continue;
		}
		
		s = new CSClause;
		s->m_literals = s1->m_literals;
		s->m_literals.insert(li);		
		if (!cs_new->addClause(s, true)) delete s;
		li = m_L->m_L.find_next(li);
	}	

	
	for (it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {	
		s = new CSClause;
		s->m_literals = s1->m_literals;
		s->m_literals.insert((*it_s)->m_literals.begin(),(*it_s)->m_literals.end());		
		if (!s->is_consistent() || !cs_new->addClause(s, true)) delete s;		
	}		
}

void CState::and_merge_negation(const Literals* ls) {

	Literal l;
	if (ls->size() == 0)
		return;
	//cout << "0"; fflush(stdout);
	if (ls->size() == 1) {		
		NEGATE(*(ls->begin()),l);
		//m_planner->m_timer->start(SIMPLIFICATION_TIMER);
		this->UP(l);			
		//m_planner->m_timer->end(SIMPLIFICATION_TIMER);
		return;
	}
	
	CSClause* newClause;
	
	Literals::const_iterator it_x;	
	static int count=0;	
	newClause = new CSClause;	
	for (it_x = ls->begin(); it_x != ls->end(); it_x++) {
		if (m_L->m_L[*it_x]== false) {			
			NEGATE(*it_x,l);			
			newClause->m_literals.insert(l);
		}
	}		
	if (newClause->m_literals.size() == 1) {
		//m_planner->m_timer->start(SIMPLIFICATION_TIMER);
		this->UP(*newClause->m_literals.begin());
		//m_planner->m_timer->end(SIMPLIFICATION_TIMER);
		delete newClause;		
	}
	else {		
		if (!this->addClause(newClause, false)) delete newClause;		
	}		
	//if (newClause->m_literals.size() == 1)
	//	simplifyCNF(this);
}


bool CState::goal_satisfied()
{  
  Literals::iterator it;
  set<Literals>::iterator it1;
  int i;
  bool inS;
  Literal l;

  if (m_L->goal_satisfied())	  
	  return true;  
  
  m_hvalue = m_L->m_gsatisfied.count();

  for (it1 = (m_L->m_planner->m_gd).begin(), i=0; it1 != (m_L->m_planner->m_gd).end(); it1++, i++) {
	  if (!m_L->m_gsatisfied[i]) {
		  if (it1->size() == 1) {
			  NEGATE(*it1->begin(), l);
			  if (m_L->m_L[l]) return false;
		  }
		  
		  if (!is_inClause(*it1))
			 return false;		
	  }
  }
    
  bool unsat = this->unsatisfiable_with(&(m_planner->m_ngd));  
  
  return unsat;
}

/*
bool CState::goal_satisfied() //This also count the number of satisfied sub-goals
{  
  Literals::iterator it;
  set<Literals>::iterator it1;
  int i;
  bool inS;
  Literal l;

  if (m_hvalue != -1)
	  return (m_hvalue == (m_L->m_planner->m_gd).size());

  if (m_L->goal_satisfied())
	  return true;   
  //cout << "CState = "; this->print();
  //cout << "State address: " << m_L;
  //if (m_L->m_allocated)
  //	  cout << " allocated" <<endl;
  //else
  //	  cout << " NOT allocated" <<endl;	


  for (it1 = (m_L->m_planner->m_gd).begin(), i=0; it1 != (m_L->m_planner->m_gd).end(); it1++, i++) {
	  if (!m_L->m_gsatisfied[i]) {
		  if (it1->size() == 1) {
			  NEGATE(*it1->begin(), l);
			  if (m_L->m_L[l]) continue;
		  }
		  
		  if (!is_inClause(*it1))
			 continue;
		  if (implies_clause(&(*it1))){
			  if (m_L->m_allocated) {
			  m_L->m_gsatisfied[i]=1;
			  if (it1->size() == 1)
			  	UP(*it1->begin());
			  //m_hvalue++;
			  }
		  }
	  }
  }

  m_hvalue = m_L->m_gsatisfied.count();
  //cout <<"m_hvalue = " << m_hvalue << endl;
  return (m_hvalue == (m_L->m_planner->m_gd).size());
}
*/

CSClause* CState::is_inClause(const Literal l) const {
	set<CSClause*>::iterator its;
	
	if (m_inS.size()) {		
		if (m_inS[l].size())
			return (*m_inS[l].begin());		
		else
			return NULL;
	}
	
	for (its = m_clauses.begin(); its != m_clauses.end(); its++) {
		if ((*its)->m_literals.count(l)) return *its;
	}
	
	return NULL;
}


bool CState::is_inClauses() const {

	if (!m_inS.size()) {
		cout << "NOT inClaused ";
		return true;
	}

	set<CSClause*>::iterator its;
	Literals::const_iterator itl;
	Literal l;
	
	for (its = m_clauses.begin(); its != m_clauses.end(); its++) {
		for (itl = (*its)->m_literals.begin(); itl != (*its)->m_literals.end(); itl++) {
			if (!m_inS[*itl].count(*its))
				return false;
		}
	}

	for (l = 0; l < m_inS.size(); l++) {
		for (its = m_inS[l].begin(); its != m_inS[l].end(); its++) {
			if (!m_clauses.count(*its)) 
				return false;
		}
	}

	return true;
}



bool CState::is_inClause(const Literals& x) const {
	Literals::const_iterator itl;
	
	for (itl = x.begin(); itl != x.end(); itl++) {
		if (is_inClause(*itl)) return true;
	}

	return false;
}


bool CState::intersect(const Literals* x) const
{
	Literals::const_iterator it_x;

	for (it_x = x->begin(); it_x != x->end(); it_x++) {
		if (m_L->m_L[*it_x])
			return true;
	}
	
	return false;
}

bool CState::includes(const Literals* x) const
{
  Literals::const_iterator it_x;
  
  for (it_x = x->begin(); it_x != x->end(); it_x++) {    
	if (!m_L->m_L[*it_x])
        return false;
  }  

  return true;
}

bool CState::will_include(const Literals* x) const
{
  Literals::const_iterator it_x;
  
  for (it_x = x->begin(); it_x != x->end(); it_x++) {    
	if (!m_L->m_L[*it_x] && !m_EF.count(*it_x))
        return false;
  }  

  return true;
}


CSClause* CState::includeClause(const CSClause* s) const {
	set<CSClause*>::iterator its;
	
	if (m_inS.size()) {
		Literals::iterator itl;		
		set<CSClause*> examined_set;
		for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {
			for (its = m_inS[*itl].begin(); its != m_inS[*itl].end(); its++) {
				if (!examined_set.count(*its)) {				
					if ((*its)->m_literals == s->m_literals) return *its;				
					examined_set.insert(*its);
				}
			}
		}
	}
	else {
		for (its = m_clauses.begin(); its != m_clauses.end(); its++) {
			if ((*its)->m_literals == s->m_literals) return *its;							
		}
	}

	return NULL;
}


CSClause* CState::subsumes(const CSClause* s) const {
	set<CSClause*>::iterator its;
	
	if (m_inS.size()) {
		Literals::iterator itl;	
		set<CSClause*> examined_set;
		for (itl = s->m_literals.begin(); itl != s->m_literals.end(); itl++) {
			for (its = m_inS[*itl].begin(); its != m_inS[*itl].end(); its++) {
				if (!examined_set.count(*its)) {				
					if ((*its)->subsumes(s)) return *its;				
					examined_set.insert(*its);
				}
			}
		}
	}
	else {
		for (its = m_clauses.begin(); its != m_clauses.end(); its++) {
			if ((*its)->subsumes(s)) return *its;
		}
	}

	return NULL;
}


bool CState::subsumes(const Literals* s) const {
	set<CSClause*>::iterator its;
	
	if (m_inS.size()) {
		Literals::iterator itl;	
		set<CSClause*> examined_set;

		for (itl = s->begin(); itl != s->end(); itl++) {	
			for (its = m_inS[*itl].begin(); its != m_inS[*itl].end(); its++) {			
				if (!examined_set.count(*its)) {				
					if ((*its)->subsumes(s)) {					
						return true;				
					}
					examined_set.insert(*its);
				}			
			}		
		}
	}
	else {
		for (its = m_clauses.begin(); its != m_clauses.end(); its++) {
			if ((*its)->subsumes(s))
				return true;
		}
	}

	return false;
}

bool CState::implies(const Action* act) const
{
  ExecList::const_iterator its;
  Literals::iterator it_x;
  Literal l;  
  
  if (act->get_execs()->size() == 1) {	  
	  for (it_x = (act->get_execs()->begin())->begin(); it_x != (act->get_execs()->begin())->end(); it_x++) {  	  
		  NEGATE(*it_x,l);
		  if (m_L->m_L[l])
			  return false;
		  if (!m_L->m_L[*it_x] && !is_inClause(*it_x))
			  return false;
	  }
  }

 // cs => x if (cs ^ -x) is unsatisfiable  
  bool unsat = unsatisfiable_with(act->get_neg_execs());  
  return unsat;
}

//Check implication only failed after check including
bool CState::implies(const Literals* x) const
{
  Literals::const_iterator it_x;
  Literal l;
  
  for (it_x = x->begin(); it_x != x->end(); it_x++) {
    
    NEGATE(*it_x,l);
	if (m_L->m_L[l])
		return false;
	if (!m_L->m_L[*it_x] && !is_inClause(*it_x))
        return false;
  }

  // cs => x if (cs ^ -x) is unsatisfiable  
  return (!solvable_assume_negation(x));  
}


bool CState::unsatisfiable_with(const set<Literals>* cnf) const
{
	Solver      S;	
    S.verbosity = 1;
	vec<Lit> lits;

	Literals nx;
	
	Literals::const_iterator it_x;
	set<CSClause*>::const_iterator it_s;
	set<Literals>::iterator it;

	/* Don't need to pass L set to the solver
	for (it_x = m_L->m_literals.begin(); it_x != m_L->m_literals.end(); it_x++) {
		readUnitClause(*it_x, S, lits);
		S.addClause(lits);		
	}			
	*/

	for (it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {
		readClause1((*it_s)->m_literals, S, lits);
		S.addClause(lits);
	}

	//Pass the cnf to the solver to check if cs & cnf is satisfiable 
	for (it = cnf->begin(); it != cnf->end(); it++) {
		if (this->intersect(&(*it)))
			continue;
		readClause1(*it, S, lits, m_L->m_L);
		if (lits.size())
			S.addClause(lits);
	}

	if (!S.simplify())
		return true;  //cs entails m_gd	

	
	//return (!S.solve());
	
	if (S.solve()) 
		return false;
	else {
		//cout << "SAT needed!";
		return true;
	}
}

// x is a conjunction, CS ^ -x unsat <=> CS |= x
bool CState::solvable_assume_negation(const Literals* x) const
{
	Solver      S;	
    S.verbosity = 1;
	vec<Lit> lits;

	Literals nx;
	Literal l;
	Literals::const_iterator it_x;
	set<CSClause*>::const_iterator it_s;
	int h = 0;	
	for (it_x = x->begin(); it_x != x->end(); it_x++) {
		if (!m_L->m_L[*it_x]) {
			NEGATE(*it_x,l);
			nx.insert(l); 
		}
	}
	
	readClause1(nx, S, lits);
    S.addClause(lits);	
	for (it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {		
		readClause1((*it_s)->m_literals, S, lits);		
		S.addClause(lits);
	}	
	if (!S.simplify())
		return false;

	//return S.solve();
	
	if (S.solve())
		return true;
	else {
		//cout << "SAT not compact";
		return false;
	}
}

bool CState::implies_clause(const Literals* x) const
{
	Solver      S;	
    S.verbosity = 1;
	vec<Lit> lits;
	
	Literal l;
	Literals::const_iterator it_x;
	set<CSClause*>::const_iterator it_s;

	for (it_x = x->begin(); it_x != x->end(); it_x++) {		
		NEGATE(*it_x,l);
		readUnitClause(l, S, lits);		
		S.addClause(lits);
	}

	for (it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {		
		readClause1((*it_s)->m_literals, S, lits);		
		S.addClause(lits);
	}	
	if (!S.simplify())
		return true;

	//return S.solve();
	
	if (S.solve())
		return false;
	else {
		//cout << "SAT not compact";
		return true;
	}
}

bool CState::entails(const Literal l) const
{
	
	if (!is_inClause(l))
		return false;
	
	Solver      S;	
    S.verbosity = 1;
	vec<Lit> lits;

	set<CSClause*>::const_iterator it_s;

	for (it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {
		readClause1((*it_s)->m_literals, S, lits);
		S.addClause(lits);
	}

	if (!S.simplify()) {
		cout <<" Unsatisfiable CState, Check!";
		return false;
	}

	Literal l1;
	NEGATE(l, l1);

	readUnitClause(l1, S, lits);
	return !S.solve(lits);
}



bool CState::solvable_assume(const Literal l) const
{
	Solver      S;	
    S.verbosity = 1;
	vec<Lit> lits;

	set<CSClause*>::const_iterator it_s;

	for (it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {
		readClause1((*it_s)->m_literals, S, lits);
		S.addClause(lits);
	}

	if (!S.simplify())
		return false;	

	Literal l1;
	NEGATE(l, l1);

	if (is_inClause(l1)) {			
		readUnitClause(l, S, lits);
		return S.solve(lits);
	}
	return true;
}


bool CState::solvable_assume(const Literals* x) const
{
	Solver      S;	
    S.verbosity = 1;
	vec<Lit> lits;

	Literals nx;
	Literal l;
	
	Literals::const_iterator it_x;
	set<CSClause*>::const_iterator it_s;

	for (it_x = x->begin(); it_x != x->end(); it_x++) {
		if (m_L->m_L[*it_x]) continue;
		NEGATE(*it_x, l);
		if (is_inClause(l)) {
			nx.insert(*it_x);
		}
	}		
	if (!nx.size())
		return true;

	for (it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {
		readClause1((*it_s)->m_literals, S, lits);
		S.addClause(lits);
	}

	if (!S.simplify())
		return false;
	
	//Check satisfiability of S assuming the set of litterals nx are true
	readClause1(nx, S, lits);    	
	//return S.solve(lits);
	
	if (S.solve(lits))
		return true;
	else {
		//cout << "SAT needed" << endl;
		return false;
	}
}

//Return the action that when executed in cs results in this cstate
const Action* CState::get_action(CState* cs)
{
	map<CState*,const Action*>::iterator it;

	it = m_in_transitions.find(cs);

	if (it != m_in_transitions.end())
		return it->second;

	return NULL;
}

//Return true if : cs x act -> this
bool CState::is_successor(CState* cs, const Action* act)
{
	map<CState*,const Action*>::iterator it;

	it = m_in_transitions.find(cs);

	if (it != m_in_transitions.end() && it->second == act)
		return true;

	return false;
}

bool CState::decides(const Literal l)
{
	Literal l1;

	if (m_L->m_L[l]) return true;
	NEGATE(l, l1);
	if (m_L->m_L[l1]) return true;
	
	if (entails(l) || entails(l1)) {
		cout << "NOT OPTIMIZE ";
		return true;
	}
	else 
		return false;

	return (entails(l) || entails(l1));
}

void CState::split(CState* cs1, CState* cs2, const Literal l)
{
	Literal l1;
	NEGATE(l, l1);
	cs1->copy(this);
	//m_planner->m_timer->start(SIMPLIFICATION_TIMER);
	cs1->UP(l);
	//m_planner->m_timer->end(SIMPLIFICATION_TIMER);
	cs1->allocate();
	cs2->copy(this);
	//m_planner->m_timer->start(SIMPLIFICATION_TIMER);
	cs2->UP(l1);
	//m_planner->m_timer->end(SIMPLIFICATION_TIMER);
	cs2->allocate();
}

int CState::hgvalue() //Can't combine with CState::hvalue()
{  
	if (m_hvalue == -1)
		m_hvalue = m_L->hgvalue();
	return m_hvalue;
}


int CState::hcvalue() 
{
  
  if (m_hvalue != -1)
	return m_hvalue;
  m_hvalue = 0;
  Literals::const_iterator itl;
  for (itl = m_planner->m_bd.begin(); itl!= m_planner->m_bd.end(); itl++)
	if(m_L->m_L[*itl])
		m_hvalue++;
  return m_hvalue;

  return m_L->m_L.count();
  set<CSClause*>::iterator it_s;
  int h = 0;
  for ( it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {
	h += (*it_s)->m_literals.size();
  }
  return h;	
}


int CState::hvalue1()
{  
	if (m_hvalue == -1) {
		set<Literals>::iterator its;
		set<CSClause*>::iterator it_s;
		Literals::iterator itl;
		int i;
		int min = -1;
		
		m_hvalue = 0;
		
		for (its = m_planner->m_gd.begin(); its != m_planner->m_gd.end(); its++) {						
			if (this->includes(&(*its))) {			
				m_hvalue += 1000;
				continue;
			}
						
			if (this->subsumes(&(*its))) {
				m_hvalue += 1000;
				continue;
			}
			
			for ( it_s = m_clauses.begin(); it_s != m_clauses.end(); it_s++) {
				if ((*it_s)->intersect(&(*its))) {
					m_hvalue += 1;
					break;
				}
			}			
		}		
	}	
	
	return m_hvalue;
}


int CState::hvalue() 
{  	
	
	//return m_L->m_gsatisfied.count();
	if (m_hvalue == -1)
		m_hvalue = m_L->m_gsatisfied.count(); 
	return m_hvalue;
       
}

int CState::size() {
	if (m_size == -1) {
		set<CSClause*>::iterator its;	
		m_size = 0;
		for (its = m_clauses.begin(); its != m_clauses.end(); its++) {		
			m_size += (*its)->m_literals.size();
		}		
  	}  
  	return m_size;
}

int CState::hash()
{
  if (m_hash == -1) {    
	set<CSClause*>::iterator its;	
	unsigned int hash = m_L->hash();
	for (its = m_clauses.begin(); its != m_clauses.end(); its++) {		
		hash += (*its)->hash();
	}	
	m_hash = hash & HASH_BITS;	
  }  

  return m_hash;
}

void CState::solution_print(Planner* planner, const int level)
{  	
	int i;	
	map<CState*,int>::iterator it;	
	const Action* act;	
	
	for (i=0; i<level; i++)	cout << "  ";
			  
	if (this->goal_satisfied()) {		
		if (planner->m_depth < this->m_depth)
			planner->m_depth = this->m_depth;
		return;
	}
	
	if (!m_successors.size()) {
		cout <<"DEAD-END, WRONG PLAN!";
		return;
	}

	act = m_successors[0]->get_action(this);
	cout << planner->m_commandNo.size() <<": " << act->get_name()<<endl;
	planner->m_commandNo.insert(pair<CState*,int>(this,planner->m_commandNo.size()));	
	
	if (!act->m_sensing) {
		it = planner->m_commandNo.find(m_successors[0]);
		if (it == planner->m_commandNo.end()) {
			m_successors[0]->m_depth = m_depth + 1;
			m_successors[0]->solution_print(planner,level);			
		}
		else {
			if (m_successors[0]->m_depth <= m_depth) {
				m_successors[0]->m_depth = m_depth + 1;
				m_successors[0]->depth_propagate(planner);
			}

			for (i=0; i < level; i++)	cout << "  ";
			cout << "GOTO " << it->second << endl;
		}
	}
	else {
		
		if (m_successors.size() < 2) {
			cout <<"WRONG TRANSITION STRUCTURE!";
			return;
		}

		cout << "  " << "IF: " << (m_planner->convert(act->lit())) << endl;
		it = planner->m_commandNo.find(m_successors[0]);
		if (it == planner->m_commandNo.end()) {
			m_successors[0]->m_depth = m_depth + 1;
			m_successors[0]->solution_print(planner,level+1);			
		}
		else {
			if (m_successors[0]->m_depth <= m_depth) {
				m_successors[0]->m_depth = m_depth + 1;
				m_successors[0]->depth_propagate(planner);
			}

			for (i=0; i <= level; i++)	cout << "  ";
			cout << "GOTO " << it->second << endl;
		}		
		
		for (i=0; i < level; i++)	cout << "  ";
		cout << "IF: NOT " << (m_planner->convert(act->lit())) << endl;
		it = planner->m_commandNo.find(m_successors[1]);
		if (it == planner->m_commandNo.end()) {
			m_successors[1]->m_depth = m_depth + 1;
			m_successors[1]->solution_print(planner,level+1);			
		}
		else {			
			if (m_successors[1]->m_depth <= m_depth) {
				m_successors[1]->m_depth = m_depth + 1;
				m_successors[1]->depth_propagate(planner);
			}

			for (i=0; i <= level; i++)	cout << "  ";
			cout << "GOTO " << it->second << endl;
		}
	}
	return;
}

void CState::solution_not_print(Planner* planner, const int level)
{  	
	int i;	
	map<CState*,int>::iterator it;	
	const Action* act;	
	
	//for (i=0; i<level; i++)	cout << "  ";

	if (this->goal_satisfied()) {		
		if (planner->m_depth < this->m_depth)
			planner->m_depth = this->m_depth;
		return;
	}
	
	if (!m_successors.size()) {
		cout <<"DEAD-END, WRONG PLAN!";
		return;
	}

	act = m_successors[0]->get_action(this);
	//cout << planner->m_commandNo.size() <<": " << act->get_name()<<endl;
	planner->m_commandNo.insert(pair<CState*,int>(this,planner->m_commandNo.size()));	
	
	if (!act->m_sensing) {
		it = planner->m_commandNo.find(m_successors[0]);
		if (it == planner->m_commandNo.end()) {
			m_successors[0]->m_depth = m_depth + 1;
			m_successors[0]->solution_not_print(planner,level);			
		}
		else {
			if (m_successors[0]->m_depth <= m_depth) {
				m_successors[0]->m_depth = m_depth + 1;
				m_successors[0]->depth_propagate(planner);
			}
			//for (i=0; i < level; i++)	cout << "  ";
			//cout << "GOTO " << it->second << endl;
		}
	}
	else {
		
		if (m_successors.size() < 2) {
			cout <<"WRONG TRANSITION STRUCTURE!";
			return;
		}

	//	cout << "  " << "IF: " << (m_planner->convert(act->lit())) << endl;
		it = planner->m_commandNo.find(m_successors[0]);
		if (it == planner->m_commandNo.end()) {
			m_successors[0]->m_depth = m_depth + 1;
			m_successors[0]->solution_not_print(planner,level+1);			
		}
		else {
			if (m_successors[0]->m_depth <= m_depth) {
				m_successors[0]->m_depth = m_depth + 1;
				m_successors[0]->depth_propagate(planner);
			}
	//		for (i=0; i <= level; i++)	cout << "  ";
	//		cout << "GOTO " << it->second << endl;
		}		
		
	//	for (i=0; i <= level; i++)	cout << "  ";
	//	cout << "IF: NOT " << (m_planner->convert(act->lit())) << endl;
		it = planner->m_commandNo.find(m_successors[1]);
		if (it == planner->m_commandNo.end()) {
			m_successors[1]->m_depth = m_depth + 1;
			m_successors[1]->solution_not_print(planner,level+1);			
		}
		else {
			if (m_successors[1]->m_depth <= m_depth) {
				m_successors[1]->m_depth = m_depth + 1;
				m_successors[1]->depth_propagate(planner);
			}
	//		for (i=0; i <= level; i++)	cout << "  ";
	//		cout << "GOTO " << it->second << endl;
		}
	}
	return;
}

void CState::depth_propagate(Planner* planner)
{
	int i;
	for (i=0; i<m_successors.size(); i++) {
		if (m_successors[i]->m_depth <= m_depth) {
			m_successors[i]->m_depth = m_depth + 1;
			m_successors[i]->depth_propagate(planner);
		}
	}	
	if (!m_successors.size() && planner->m_depth < m_depth)
			planner->m_depth = m_depth;
}


void CState::print() const
{
  bool comma = false, inClause= true;
  Literals::iterator it_l;
  set<CSClause*>::const_iterator it;    

  cout << "L = "; m_L->print(); cout <<endl; fflush(stdout);
  cout << "<{";
  for (it = m_clauses.begin(); it != m_clauses.end(); it++) {
    if (comma) {
      cout << ",";
    }
    else 
      comma = true;
    //(*it)->print();
	cout << "{"; m_planner->print((*it)->m_literals); cout << "}";
	//cout<<*it;
  }
  cout << "}>" <<endl; fflush(stdout);
}


void CState::print_summary() const
{
  set<CSClause*>::const_iterator it;
  int litcount = 0;
  
  for (it = m_clauses.begin(); it != m_clauses.end(); it++) {	  
		  litcount += (*it)->m_literals.size();	  
  }

  cout <<"Number of unit clauses: " << m_L->m_L.count() <<endl;
  cout <<"Number of other clauses: " << m_clauses.size()<<endl;
  cout <<"Number of literals in clauses: " << litcount <<endl;
  cout <<"Average clause size: "<< (float) litcount/m_clauses.size()<<endl;
}