#ifndef CNFSTATE_H
#define CNFSTATE_H
#include "define.h"
#include "cnfsolver.h"

class Planner;
class Action;

/***********************************************************************
 Types -- Domain Description
************************************************************************/
class Effect {
 protected:
  Literals m_head;
  Literals m_body;
  bool m_nondet;
  Literals m_neg_body; // negation of the body
 public:
  Effect();
  ~Effect();

  void set_body(Literals bd);
  void set_head(Literals hd);
  bool nondet() const;
  const Literals* get_body() const;
  const Literals* get_neg_body() const;
  const Literals* get_head() const;
};

typedef list<Effect> EffectList;
typedef list<Literals> ExecList;

typedef Effect StaticLaw;
typedef list<StaticLaw> StaticLaws;

class State
{
public:
  boost::dynamic_bitset<> m_L;
  boost::dynamic_bitset<> m_gsatisfied;
  int m_gchecked;
  int m_hvalue;     // heuristic value
  int m_hash;
  bool m_allocated;
  

public:
  const Planner* m_planner;	
  State(const Planner* planner);
  ~State();

  /* operators */
  void clear();
  int operator<(const State&) const;
  bool operator==(const State&) const;
  State operator=(const State& s);
  bool intersect(const Literals* x) const;
  bool intersect(const State* s) const;
  bool includes(const Literals* x) const;
  bool includes(const State* s) const;    

  /* main functions */
  void closure();

  /* other functions */
  bool is_consistent() const;
  bool satisfies(const Literals* x);
  bool goal_satisfied();
  int hgvalue();
  int hvalue();
  int hash();
  void print() const;
};

class StateComp
{
 public:
  int operator()(const State* s1, const State* s2) const {
    return *s1 < *s2;
  }
};

typedef set<State*,StateComp> StateTable;

class CSClause
{
public:
  Literals m_literals;
  int m_hash;  
  bool m_allocated;

public:  
  CSClause();
  ~CSClause();

  /* operators */
  void clear();
  int operator<(const CSClause&) const;
  bool operator==(const CSClause&) const;
  CSClause operator=(const CSClause& s);
  bool is_consistent() const;
  bool intersect(const Literals* x) const;
  bool intersect(const CSClause* s) const;
  bool includes(const Literals* x) const;
  bool includes(const CSClause* s) const;  
  bool subsumes(const CSClause* s) const;
  bool subsumes(const Literals* s) const;

  /* main functions */
  void closure();

  /* other functions */
  int hash();
  void print() const;
};

class ClauseComp
{
 public:
  int operator()(const CSClause* s1, const CSClause* s2) const {
    return *s1 < *s2;
  }
};

typedef set<CSClause*, ClauseComp> ClauseTable;


class CState {
  friend class PriotizedCState;
  friend class CStateComp;
 public:
  State* m_L;
  Literals m_EF;
  Literals m_nondet_EF;
  set<CSClause*> m_clauses;
  vector<set<CSClause*> > m_inS;
  int m_hvalue;
  int m_size;
  int m_hash;
  bool m_allocated;
  
  short m_status;    
  map<CState*, const Action*> m_in_transitions; //Incomming transitions
  int m_active_in_transitions; // Incoming transitions not be goal_propagated
  vector<CState*> m_successors; //successor nodes obtained by sensing  
  int m_depth;   // length of the plan  

protected:  
  int m_plan_length;   // length of the plan
  const Action* m_action;
  const CState* m_previous_cstate;
public:
  Planner* m_planner;
  /* constructor and destructor */
  //CState(const Planner* planner);
  CState();	
  ~CState();

  void clear();
  bool operator<(const CState& right) const;
  //CState operator=(const CState& s);
  void copy(const CState* cs);
  void preprocess();
  CSClause* resolvent(const CSClause* s, const CSClause* s1, const Literal l, const Literal l1);
  bool checkSubsumption(const CSClause* s, set<CSClause*>* P, vector<CSClause*>* unchecked);
  bool removeSubsumed(const CSClause* s);
  
  void removeClause(CSClause* s);  
  void inClause(CSClause* s);
  int  Unit_Extract();
  void UP();
  void UP(const Literal pl);
  void UP(const Literals& st);
  void outClause(CSClause* s);
  bool addClause(CSClause* s, const bool checkSubsumed);
  void allocate();  
  void inClauses();
  void and_merge(const Literal l);
  void and_merge(const Literals* ls);  
  void or_merge(CState* cs,CState* cs_new );
  void or_merge(Literal l, CState* cs_new) const;
  void or_merge(CSClause* s1, CState* cs_new) const;
  void and_merge_negation(const Literals* ls);

  const Action* get_action(CState* cs);
  bool is_successor(CState* cs, const Action* act);
  

  bool goal_satisfied();  
  bool unsatisfiable_with(const set<Literals>* cnf) const;
  CSClause* is_inClause(const Literal l) const;
  bool is_inClause(const Literals& x) const;
  bool is_inClauses() const;
  bool intersect(const Literals* x) const;
  bool includes(const Literals* x) const;  
  bool will_include(const Literals* x) const;
  CSClause* includeClause(const CSClause* s) const;
  CSClause* subsumes(const CSClause* s) const;
  bool subsumes(const Literals* s) const;
  bool implies(const Action* act) const;
  bool implies(const Literals* x) const;    
  bool implies_clause(const Literals* x) const;

  bool solvable_assume_negation(const Literals* x) const;  
  bool entails(const Literal l) const;
  bool solvable_assume(const Literal l) const;
  bool solvable_assume(const Literals* x) const;  
  bool decides(const Literal l);
  void split(CState* cs1, CState* cs2, const Literal l);

  int hgvalue();
  int hcvalue();
  int hvalue1();
  int hvalue();
  int size();
  int hash();

  void solution_print(Planner* planner, const int level);
  void solution_not_print(Planner* planner, const int level);
  void depth_propagate(Planner* planner); 

  /* other functions */
  void print() const;
  void print_summary() const;  
};

class CStateComp
{
 public:
  int operator()(const CState* cs1, const CState* cs2) const {
    return *cs1 < *cs2;
  }
};

typedef set<CState*,CStateComp> CStateTable;

// Friend class with function object for comparison of cnodes
class PriotizedCState0 {
public :
  int operator()(CState* cs1, CState* cs2) const
  {
	//  if (cs1->hvalue1() == cs2->hvalue1()) {
		//if (cs1->get_plan_length() == cs2->get_plan_length())
	//		return cs1->hcvalue() < cs2->hcvalue();
		//return cs1->get_plan_length() > cs2->get_plan_length();
	  //}

	  //if (cs1->hvalue() == cs2->hvalue())
		//  return cs1->hcvalue() < cs2->hcvalue();
	//  return cs1->hvalue() < cs2->hvalue();		

	//if (cs1->hgvalue() == cs2->hgvalue())
	//	  return cs1->hcvalue() < cs2->hcvalue();
	//return cs1->hgvalue() < cs2->hgvalue();

		  //return cs1->hvalue1() < cs2->hvalue1();
	//return cs1->hgvalue() < cs2->hgvalue();
/*     //This solve coins_21 problem
	if (cs1->hvalue() == cs2->hvalue()) {
		if (cs1->hcvalue() == cs2->hcvalue())
			return cs1->m_clauses.size() > cs2->m_clauses.size();
		return cs1->hcvalue() < cs2->hcvalue();
	}
		//return cs1->m_clauses.size() > cs2->m_clauses.size();
	return cs1->hvalue() < cs2->hvalue();	
*/
//return cs1->hvalue() < cs2->hvalue();
	if (cs1->hvalue() == cs2->hvalue()) {
		return cs1->size() >= cs2->size();
	//	if (cs1->m_clauses.size() == cs2->m_clauses.size())
	//		return cs1->hcvalue() < cs2->hcvalue();
	//	return cs1->m_clauses.size() > cs2->m_clauses.size();
	//	return cs1->m_L->m_gsatisfied.count() < cs2->m_L->m_gsatisfied.count();
	}
	return cs1->hvalue() < cs2->hvalue();
  }
};

class PriotizedCState3 {
public :
  int operator()(CState* cs1, CState* cs2) const
  {		  
	if (cs1->hvalue() == cs2->hvalue()) {
		if (cs1->hcvalue() == cs2->hcvalue()) {
			if (cs1->m_L->m_L.count() == cs2->m_L->m_L.count())
				return cs1->m_clauses.size() > cs2->m_clauses.size();
			return cs1->m_L->m_L.count() < cs2->m_L->m_L.count();
		}
		return cs1->hcvalue() < cs2->hcvalue();
	}		
	return cs1->hvalue() < cs2->hvalue();
  }
};

class PriotizedCState { //This solves sortnet
public :
  int operator()(CState* cs1, CState* cs2) const
  {	    
	if (cs1->hvalue() == cs2->hvalue()) {
		//if (cs1->m_L->m_L.count() == cs2->m_L->m_L.count()) {				
		//	if (cs1->m_clauses.size() == cs2->m_clauses.size())
		//		return cs1->size() > cs2->size();
			//return cs1->m_clauses.size() > cs2->m_clauses.size();		
		//}
		return cs1->m_L->m_L.count() < cs2->m_L->m_L.count();
	}		
	return cs1->hvalue() < cs2->hvalue();
  }
};


class PriotizedCState1 {
public :
  int operator()(CState* cs1, CState* cs2) const
  {	    
	//if (cs1->hvalue() == cs2->hvalue()) {
	//	if (cs1->m_L->m_L.count() == cs2->m_L->m_L.count())
	//		if (cs1->m_clauses.size() == cs2->m_clauses.size())
	//			return cs1->size() > cs2->size();
	//	return cs1->m_L->m_L.count() < cs2->m_L->m_L.count();
	//}
	return cs1->hvalue() < cs2->hvalue();	
  }
};

class PriotizedCState2 {
public :
  int operator()(CState* cs1, CState* cs2) const
  {    
	if (cs1->m_L->m_L.count() == cs2->m_L->m_L.count()) 
		return cs1->size() > cs2->size();
	return cs1->m_L->m_L.count() < cs2->m_L->m_L.count();	
  }
};

//Solution 1

/* Results in the paper.
class PriotizedCState {
public :
  int operator()(CState* cs1, CState* cs2) const
  {		  
	if (cs1->hvalue() == cs2->hvalue()) {
		if (cs1->hcvalue() == cs2->hcvalue()) {
			if (cs1->m_L->m_L.count() == cs2->m_L->m_L.count())
				return cs1->m_clauses.size() > cs2->m_clauses.size();
			return cs1->m_L->m_L.count() < cs2->m_L->m_L.count();
		}
		return cs1->hcvalue() < cs2->hcvalue();
	}		
	return cs1->hvalue() < cs2->hvalue();
  }
};

class PriotizedCState1 {
public :
  int operator()(CState* cs1, CState* cs2) const
  {		  
	if (cs1->hvalue() == cs2->hvalue()) {
		if (cs1->m_L->m_L.count() == cs2->m_L->m_L.count()) {
			if (cs1->hcvalue() == cs2->hcvalue())
				return cs1->m_clauses.size() > cs2->m_clauses.size();
			return cs1->hcvalue() < cs2->hcvalue();
		}
		return cs1->m_L->m_L.count() < cs2->m_L->m_L.count();
	}		
	return cs1->hvalue() < cs2->hvalue();
  }
};
*/


//Solution 2 (better than sol 1)

/*
class PriotizedCState {
public :
  int operator()(CState* cs1, CState* cs2) const
  {		  
	if (cs1->hvalue() == cs2->hvalue()) {
		if (cs1->m_L->m_L.count() == cs2->m_L->m_L.count()) {
			if (cs1->hcvalue() == cs2->hcvalue())
				return cs1->m_clauses.size() > cs2->m_clauses.size();
			return cs1->hcvalue() < cs2->hcvalue();
		}
		return cs1->m_L->m_L.count() < cs2->m_L->m_L.count();
	}		
	return cs1->hvalue() < cs2->hvalue();
  }
};

*/

//Solution 3 same as solution 2 for the problems on paper
/*
class PriotizedCState {
public :
  int operator()(CState* cs1, CState* cs2) const
  {		  
	if (cs1->hvalue() == cs2->hvalue()) {
		if (cs1->m_L->m_L.count() == cs2->m_L->m_L.count()) {
			//if (cs1->hcvalue() == cs2->hcvalue())
				return cs1->m_clauses.size() > cs2->m_clauses.size();
			//return cs1->hcvalue() < cs2->hcvalue();
		}
		return cs1->m_L->m_L.count() < cs2->m_L->m_L.count();
	}		
	return cs1->hvalue() < cs2->hvalue();
  }
};
*/



typedef priority_queue<CState*,vector<CState*>,PriotizedCState> PriorityCStateQueue;
typedef priority_queue<CState*,vector<CState*>,PriotizedCState1> PriorityCStateQueue1;
typedef priority_queue<CState*,vector<CState*>,PriotizedCState2> PriorityCStateQueue2;
#endif
