// planner.cpp: implementation of the planner class.
//////////////////////////////////////////////////////////////////////
#include "planner.h"

clock_t t_start, t_now;
bool giveup;
double timeLimit;

/*************************************************************************
 * Planner Implementation
 *************************************************************************/
// Constructor & destructor 
Planner::Planner(const Reader* reader, Timer* timer)
{
  m_timer = timer;
  m_reader = reader;

  m_semantics = PC;
  m_task = SEARCH;
  m_algorithm = GREEDY;
  m_detect_essential = true;
  m_output_decisive = false;  
}

Planner::~Planner()
{
 
}

bool Planner::main()
{

#ifdef PRINT_DOMAIN
  m_reader->print();
  exit(0);
#endif
   
  Solver      S;
  S.verbosity = 1;

  // build the domain
  cout << "Building domain..."; fflush(stdout);
  m_timer->start(BUILD_TIMER);
  m_states.resize(HASH_SIZE);
  m_clauses.resize(HASH_SIZE);
  m_cstates.resize(HASH_SIZE);

  build_domain();   
  build_goal();        
  build_init();
  
  m_timer->end(BUILD_TIMER);

  cout << "done." << endl << endl;
  cout << "Number of literals: " << m_literals.size() << endl;
  cout << "Number of goals: " << m_gd.size() << endl;
  fflush(stdout);
  
  // print domain
#ifdef PRINT_INTERNAL_DOMAIN
  print_interal_domain();
  cout << endl;
#endif

#ifdef PRINT_SUMMARY
  cout << endl;
  print_summary();
#endif
  // search for a plan or execute a sequence of actions
  switch (m_task) {
  case SEARCH:
    m_timer->start(SEARCH_TIMER);
    search();
    m_timer->end(SEARCH_TIMER);
    break;
  case DOPLAN:
    execute();
    break;
  default:
    return false;
  }

  cout << endl;
  print_statistics(); 

  return true;
};

/*

  StringList x;
  unsigned int i;
  string f; // fluent
  StringList::const_iterator itf;
  typedef map<string,Literal>::value_type value_type;
*/

bool Planner::build_domain()
{
  int i;
  StringList::const_iterator itf;
  typedef map<string,Literal>::value_type value_type;
  PropositionList::const_iterator it_prop;
  Action* p_act;
  StaticLaw stat;
  Literals hd, bd;

  // build fluent literals;
  i = 0;
  for (itf = m_reader->m_fluents.begin(); 
       itf != m_reader->m_fluents.end(); itf++) { 
    m_map.insert(value_type(*itf,i));
    m_literals.insert(i++);
    m_map.insert(value_type(NEGATION_SYMBOL+*itf,i));
    m_literals.insert(i++);    
  }
  
  // build action list
  for (it_prop = m_reader->m_propositions.begin(); 
       it_prop != m_reader->m_propositions.end(); it_prop++) {    
    hd = ground((*it_prop).get_effect());
    bd = ground((*it_prop).get_precondition());
    m_bd.insert(bd.begin(), bd.end());
    switch ((*it_prop).n_type) {
    case STATIC:
      stat.set_head(hd);
      stat.set_body(bd);
      m_statics.push_back(stat);
      break;
    case DYNAMIC:
      p_act = add_action((*it_prop).act_name);
      p_act->add_effect(hd,bd);
      break;
    case EXECUTABILITY:
      p_act = add_action((*it_prop).act_name);
      p_act->add_exec(bd);
      break;
    case IMPOSSIBILITY:
      p_act = add_action((*it_prop).act_name);
      p_act->add_imposs(bd);
      break;
    default:
      break;
    }
  }

  //THRESHOLD = ( (m_literals.size() > 80) ? m_literals.size()/10 : 8);
  //if (THRESHOLD > 16) THRESHOLD = 16;
  //THRESHOLD = 10000000000;	
  THRESHOLD = m_literals.size();

  return true;
}

//NEW TEST


bool Planner::build_goal()
{
  StringList2::const_iterator it;
  set<Literals>::const_iterator its;
  Literals::const_iterator itl;
  Literals st;
  Literal l;   

  // build the goal
  for (it = (m_reader->m_gd).begin(); it != (m_reader->m_gd).end(); it++) {
    st = ground(&(*it));       
    m_gd.insert(st);
  }
 
  if (!simplifyCNF(&m_gd)) {	  
	  cout << "The goal:" <<endl;
	  set<Literals>::iterator itg;
	  for (itg = m_gd.begin(); itg != m_gd.end(); itg++) {
		print(*itg);
		cout << endl;
	  }
	  cout << "is not satisfiable";
	  return false;  
  }
  
  for (its = m_gd.begin(); its != m_gd.end(); its++) {
	m_goal.insert(its->begin(),its->end());
	if (its->size() == 1)
		m_gd_units.insert(*its->begin());
	else {
		st.clear();
		for (itl = its->begin(); itl != its->end(); itl++) {
			NEGATE(*itl, l);
			st.insert(l);
		}
		m_gd_neg_clauses.insert(st);
	}
	//cout << "{"; print(*its); cout << "}; ";
  }

  getNegationCNF(&m_gd, &m_ngd);
  simplifyCNF(&m_ngd);  
  return true;
}

bool Planner::build_init()
{  
  list<boost::dynamic_bitset<> > init;
  list<boost::dynamic_bitset<> >::const_iterator it;        
  Literals st;
  CSClause *s;
  
  //print(&m_reader->m_init);

  unsigned long cnf_cost = CNF_cost(&m_reader->m_init);
  unsigned long dnf_cost = DNF_cost(&m_reader->m_init);  
  cout<<"CNF Cost: "<<(long) cnf_cost<<", DNF Cost: "<<(long) dnf_cost<<endl; fflush(stdout);
  clock_t t_start0;
  t_start0 = clock();

  if (!CNF(&m_reader->m_init, init)) {
		cout<<endl<<"Initial information is inconsistent!"<<endl;
		exit(1);
  }		  
  m_init_cstate.m_planner = this;
  m_init_cstate.m_L = new State(this);  
  for (it = init.begin(); it != init.end(); it++) {	  
	  if (it->count() == 1) {		  
		  m_init_cstate.m_L->m_L[it->find_first()] = 1;
	  }
	  else {	    
		s = new CSClause;
		insert(s->m_literals,*it);
		if (s->is_consistent()) {
			s = add_clause(s);
			m_init_cstate.m_clauses.insert(s);
		}
		else {
			delete s;		
		}
	  }
  }
  
  cout <<endl<< "Initial cstate before simplification: "<<endl; m_init_cstate.print_summary();  
  cout << "m_init_cstate.m_clauses.size() = "<< m_init_cstate.m_clauses.size()<<endl; 
    
  if (!simplifyCNF(&m_init_cstate)) {
	  cout <<"Inconsistent Initial Information!"<<endl;
	  exit(0);
  }
  //m_init_cstate.preprocess();
  m_init_cstate.inClauses();
  
  cout << "Initial cstate after simplification: "<<endl; m_init_cstate.print_summary(); 
  cout.setf(ios::fixed, ios::floatfield);
  cout.precision(3);	
  cout<<"Initial Computation Time: "<< (clock() - t_start0)/(double)CLOCKS_PER_SEC; fflush(stdout);  
  //m_init_cstate.print();

  m_init_cstate.allocate();
  //m_isize = m_init_cstate.size() + m_init_cstate.m_L->m_L.count();
  m_isize = m_init_cstate.size();
  m_size = m_isize;

  return true;
}

void Planner::print(const Node* node) {
	int i,j;	
	bool subsumed;
	vector<Literals> f, lf, rf;	
	vector<Literals>::iterator itl, itr;
	Literal l, l1;

	Literals st;
	
	if (!node) {
		cout <<"NULL, ";
		return;
	}
	//cout << node->label <<", ";

	if (node->label == "|" || node->label == "&") {			
		if (node->label == "|")
			cout<<"(v "<<CNF_cost(node)<<" "<<DNF_cost(node)<<")";
		else
			cout<<"(^ "<<CNF_cost(node)<<" "<<DNF_cost(node)<<")";
		print(node->left);
		print(node->right);		
	}
	else
		cout <<"L ";
}

unsigned long Planner::DNF_cost(const Node* node) {
	if (node->label == "|") {
		return DNF_cost(node->left) + DNF_cost(node->right);
	}
	if (node->label == "&") {
		return DNF_cost(node->left) * DNF_cost(node->right);
	}
	return 1;
}

unsigned long Planner::CNF_cost(const Node* node) {
	if (node->label == "|") {
		return CNF_cost(node->left) * CNF_cost(node->right);
	}
	if (node->label == "&") {
		return CNF_cost(node->left) + CNF_cost(node->right);
	}	
	return 1;
}	

bool Planner::timeout() {
	
	return false;

	if (giveup)
		return true;
	if ( (clock() - t_start)/(double)CLOCKS_PER_SEC > timeLimit) {
		giveup = true;
		return true;
	}
	return false;
}

void Planner::print(const list<boost::dynamic_bitset<> >& f) {
	list<boost::dynamic_bitset<> >::const_iterator it;
	cout <<"<";
	for (it = f.begin(); it != f.end(); it++)
			cout<<"{"<<*it<<"}";
	cout <<">";
}


bool Planner::CNF(const Node* node, list<boost::dynamic_bitset<> >& f) {
	int i,j;	
	bool leftOK, rightOK;
	list<boost::dynamic_bitset<> > lf, rf;
	list<boost::dynamic_bitset<> >::iterator itl, itr;
	Literal l, l1;
	boost::dynamic_bitset<> st;
		
	if (timeout()) return false;

	//cout << node->label <<", ";

	if (node->label == "&") {
		leftOK = CNF(node->left, f);
		if (timeout()) return false;

		if (!leftOK) return false;			
		
		rightOK = CNF(node->right, rf);
		if (timeout()) return false;

		if (!rightOK) {			
			return false;
		}
		
		if (annul(f, rf)) {
			f.clear();
			return false;
		}	
	
		f.merge(rf);	
		f.unique();		

		return true;
	}	

	if (node->label == "|") {

		leftOK = CNF(node->left, lf);
		if (timeout()) return false;
		if (!leftOK) {
			rightOK = CNF(node->right, f);		
			return rightOK;
		}

		rightOK = CNF(node->right, rf);
		if (timeout()) return false;		
		if (!rightOK) {
			f = lf;			
			return leftOK;
		}
		
		if (!lf.size()) {
			f = rf;			
			return true;
		}
		else if (!rf.size()) {
			f = lf;			
			return true;
		}
		else {			
			merge(lf, rf, f);
		}					

		return true;
	}
	
	st.resize(m_literals.size());
	st[ground(node->label)] = 1;	
	f.push_back(st);
	return true;
}


bool Planner::DNF(const Node* node, list<boost::dynamic_bitset<> >& f) {
	int i,j;	
	bool leftOK, rightOK;
	list<boost::dynamic_bitset<> > lf, rf;
	list<boost::dynamic_bitset<> >::iterator itl, itr;
	Literal l, l1;
	boost::dynamic_bitset<> st;
		
	if (timeout()) return false;

	if (node->label == "|") {		
		leftOK = DNF(node->left, f);
		if (timeout()) return false;

		if (!leftOK) {			
			if (!(rightOK = DNF(node->right, f)))
				return false;
			return true;
		}
		
		rightOK = DNF(node->right, rf);
		if (timeout()) return false;

		if (!rightOK) {			
			return true;
		}
		
		if (annul(f, rf)) {
			f.clear();
			return true;
		}

		f.merge(rf);
		f.unique();

		return true;
	}	

	if (node->label == "&") {		

		leftOK = DNF(node->left, lf);
		if (timeout()) return false;
		if (!leftOK) return false;	

		rightOK = DNF(node->right, rf);
		if (timeout()) return false;		
		if (!rightOK) return false;
		
		if (!lf.size()) {
			f = rf;
			return true;
		}
		else if (!rf.size()) {
			f = lf;
			return true;
		}
		else {			
			merge(lf, rf, f);			
		}				
		if (!f.size())
			return false;

		return true;
	}
	
	st.resize(m_literals.size());
	st[ground(node->label)] = 1;	
	f.push_back(st);
	return true;
}

bool Planner::is_consistent(const boost::dynamic_bitset<>& s) {
	int i;

	for (i = 0; i < s.size()-1; i += 2) {
		if (s[i] && s[i+1]) return false;
	}
	return true;
}


bool Planner::annul(const list<boost::dynamic_bitset<> >& lf, const list<boost::dynamic_bitset<> >& rf) {
	
	Literals f;
	Literal l, l1;
	list<boost::dynamic_bitset<> >::const_iterator it;
	int i, j;
	
	for (it = rf.begin(); it != rf.end(); it++) {
		if (it->count() == 1)
			f.insert(it->find_first());
	}
	
	if (!f.size())
		return false;

	for (it = lf.begin(); it != lf.end(); it++) {
		if (it->count() == 1) {
			NEGATE(it->find_first(), l);
			if (f.count(l))
				return true;
		}
	}
	
	return false;
}

bool Planner::intersect(const boost::dynamic_bitset<>& S, const Literals& x) {
	Literals::const_iterator it;
	for (it = x.begin(); it != x.end(); it++) {
		if (S[*it] == 1)
			return true;
	}
	return false;
}

bool Planner::insert(boost::dynamic_bitset<>& S, const Literals& x) {
	Literals::const_iterator it;
	bool flip = false;

	for (it = x.begin(); it != x.end(); it++) {
		if (!S[*it]) {
			S[*it] = 1;
			if (!flip) flip =true;
		}
	}
	return flip;
}


bool Planner::insert(Literals& L, const boost::dynamic_bitset<>& S) {
	int i;

	i = S.find_first();
	while (i < S.size()) {
		L.insert(i);
		i = S.find_next(i);
	}
}


void Planner::add_clause(list<boost::dynamic_bitset<> >& cs, const boost::dynamic_bitset<>& st) {	
	list<boost::dynamic_bitset<> >::iterator its;

	if (!is_consistent(st))	return;
	
	for (its = cs.begin(); its != cs.end(); its++) {
		if (its->is_subset_of(st))	return;
	}
	
	for (its = cs.begin(); its != cs.end(); ) {		
		if (st.is_proper_subset_of(*its))
			cs.erase(its);
		else
			its++;
	}

	cs.push_back(st);
}

void Planner::merge(list<boost::dynamic_bitset<> >& lf, list<boost::dynamic_bitset<> >& rf, list<boost::dynamic_bitset<> >& f) {
	list<boost::dynamic_bitset<> >::iterator its, its1, itsn;
	boost::dynamic_bitset<> st;
	bool include;	

	
	for (its = lf.begin(); its != lf.end(); ) {		
		itsn= its;
		itsn++;
		include = false;
		for (its1 = rf.begin(); its1 != rf.end() && !include; its1++) {			
			if (its1->is_subset_of(*its)) {				
				f.push_back(*its);
				if (its1->count() == its->count()) rf.erase(its1);				
				lf.erase(its);				
				include = true;				
				break;
			}
			if (include) break;			
		}				
		its = itsn;		
	}		
	

	for (its = rf.begin(); its != rf.end(); ) {
		itsn= its;
		itsn++;
		include = false;
		for (its1 = lf.begin(); its1 != lf.end() && !include; its1++) {
			if (its1->is_proper_subset_of(*its)) {
				f.push_back(*its);				
				rf.erase(its);
				include = true;
				break;
			}
			if (include) break;
		}						
		its = itsn;
	}
	

	for (its = rf.begin(); its != rf.end(); its++) {
		for (its1 = lf.begin(); its1 != lf.end(); its1++) {			
			st = *its | *its1;
			if (is_consistent(st)) f.push_back(st);			
		}
	}	
	f.sort();
	f.unique();	
}

/*================= End Initial data ===========================*/

// grounding functions
Literals Planner::ground(const StringList* x) const
{
  StringList::iterator it;
  Literals y;

  for (it = x->begin(); it != x->end(); it++) {
    y.insert(ground(*it));
  }

  return y;
}

Literal Planner::ground(const string& x) const
{
  map<string,Literal>::const_iterator p = m_map.find(x);

  if (p != m_map.end()) {
    return (p->second);
  }

  cout << "ERROR: Literal " << x << " is undeclared." << endl;
  cout << "Check the fluent declaration." << endl << endl;
  
  exit(1);
}

StringList Planner::convert(const Literals& x) const
{
  StringList y;
  Literals::iterator it;

  for (it = x.begin(); it != x.end(); it++) {
    y.insert(convert(*it));
  }

  return y;
}

string Planner::convert(const Literal& x) const
{
  unsigned short int i = x / 2;
  StringList::iterator it;

  if (i >= m_reader->m_fluents.size())
    return NULL;

  it = m_reader->m_fluents.begin();

  for (i = 0; i < x/2; i++) {
    it++;
  }

  if (x % 2 == 0)
    return *it;
  
  return (NEGATION_SYMBOL + (*it));
}


/* printing functions */
void Planner::print(const Literal& l) const
{
  //cout << convert(l);
  l % 2 == 0 ? (cout<<l/2) : cout<<"-"<<(l/2);
}

void Planner::print(const Literals& x) const
{
  Literals::iterator it;
  bool comma = false;
  for (it = x.begin(); it != x.end(); it++) {
    if (comma) 
      cout << ",";
    print(*it);
    comma = true;
  }
}

void Planner::print_interal_domain() const
{
  ActionList::const_iterator it;
  
  cout << "DOMAIN DESCRIPTION" << endl;
  cout << "----------------------------" << endl;
  for (it = m_actions.begin();it != m_actions.end(); it++) {
    it->print();
    cout << endl;    
  }

  StaticLaws::const_iterator its;
  
  cout << "Static Laws: " << endl;

  for (its = m_statics.begin(); its != m_statics.end(); ++its) {
    cout << "    ";
    print(*its->get_head());
    cout << " <- ";
    print(*its->get_body());
    cout << endl;	  
  }

  // print init cstate
  cout <<  "INIT" << endl;
  cout <<  "----------------------------" << endl;
  m_init_cstate.print();
  cout << endl;

  // print goal state
  cout <<  "GOAL " << endl;
  cout <<  "----------------------------" << endl;
  print(m_goal);
  cout << endl;  

  // print statistics
  StringList2::const_iterator itst;
  unsigned int i = 0;

  cout <<  "STATISTICS " << endl;
  cout <<  "----------------------------" << endl;
  cout << "Total actions: " << m_actions.size() << endl;
  cout << "Total fluents: " << m_literals.size() / 2 << endl;
  /*
  cout << "Unknown fluents: " << endl;
  for (itst = m_reader->m_init.begin(); 
       itst != m_reader->m_init.end(); itst++) {
    cout << "\tState " << i++ << ": ";
    cout << m_literals.size()/2 - (*itst).size();
    cout << endl;
  }
  */
  cout << endl;
}


void Planner::print_summary() const
{
  cout <<  "SUMMARY" << endl;
  cout <<  "---------------------" << endl;
  cout <<  "Number of fluents: " << m_literals.size()/2 << endl;
  cout <<  "Number of actions: " << m_actions.size() << endl;
  cout <<  "Number of static laws: " << m_statics.size() << endl;
}

void Planner::print_statistics() const
{
  double total = m_timer->time(READ_TIMER) + 
    m_timer->time(BUILD_TIMER) + m_timer->time(SEARCH_TIMER);

  cout << "STATISTICS" << endl;
  cout << "---------------------" << endl;
  printf("Total time: %.3f (sec) \n",total);
#ifdef PRINT_TIME_DETAILS
  printf("  Reading: %.3f (sec) [%.2f %%]\n", m_timer->time(READ_TIMER), 100.0 * m_timer->time(READ_TIMER) / total);
  printf("  Preprocessing: %.3f (sec) [%.2f %%]\n", m_timer->time(BUILD_TIMER), 100.0 * m_timer->time(BUILD_TIMER)/total);
  printf("  Search: %.3f (sec) [%.2f %%]\n", m_timer->time(SEARCH_TIMER), 100.0 * m_timer->time(SEARCH_TIMER)/total);  
  //printf("  Merging: %.3f (sec) [%.2f %%]\n", m_timer->time(MERGE_TIMER), 100.0 * m_timer->time(MERGE_TIMER)/total);    
  //printf("  CNF SIMPLIFY: %.3f (sec) [%.2f %%]\n", m_timer->time(SIMPLIFICATION_TIMER), 100.0 * m_timer->time(SIMPLIFICATION_TIMER)/total);
  //printf("  SAT_CHECK: %.3f (sec) [%.2f %%]\n", m_timer->time(SAT_CHECKING_TIMER), 100.0 * m_timer->time(SAT_CHECKING_TIMER)/total);
  //printf("  ACTION_EXECUTE: %.3f (sec) [%.2f %%]\n", m_timer->time(ACTION_EXECUTE_TIMER), 100.0 * m_timer->time(ACTION_EXECUTE_TIMER)/total);
  //printf("  MERGE: %.3f (sec) [%.2f %%]\n", m_timer->time(MERGE_TIMER), 100.0 * m_timer->time(MERGE_TIMER)/total);

#endif
  int i, states = 0, clauses =0, cstates = 0;
  for (i=0; i<HASH_SIZE; i++) {
	  states += m_states[i].size();
	  clauses += m_clauses[i].size();
	  cstates += m_cstates[i].size();
  }

  //printf("Total states allocated: %d\n", m_states.size());
  //printf("Total cstate(s): %d\n", m_cstates.size());
  printf("Total states allocated: %d\n", states);
  printf("Total clauses allocated: %d\n", clauses);
  printf("Total cstate(s): %d\n", cstates);
  printf("Total cstate(s) expanded: %d\n", m_expanded);
  //printf("Total cstate(s) expanded: %d, clauses: %d\n", m_expanded, m_expanded_clauses);
  //printf("Avg # of clauses expanded %.2f\n", double (double (m_expanded_clauses)/ double(m_expanded)));  
  //printf("Avg size # of expanded clauses %.2f\n", double (double (m_expanded_sum_clause_size)/ double(m_expanded_clauses)));  
  //printf("Avg number # of unknown antecedents %.10f\n", double (double (m_unknown_antecedents_sum)/ double(m_exec_actions)));  
  //printf("Initial cstate size: %d\n", m_isize);
  //printf("Average cstate size: %.2f\n", float (m_size/cstates));
}

bool Planner::execute()
{
  RawPlan::const_iterator it;
  int n_count = 0;
  CState *cs;
  const Action *act;
  map<string,int>::iterator it1;

  cs = &m_init_cstate;

  //cout << endl << "STEP " << n_count++ << endl;
  cout << "CState: ";  
  cs->print();
  if (cs->goal_satisfied())
    cout << " (GOAL)";
  
  cout << endl;

  for (it = m_plan.begin(); it != m_plan.end(); ++it) {
    it1 = m_action_map.find(*it);
    if (it1 == m_action_map.end()) {
      cout << endl << "ERROR: Could not find action " << *it 
	   << " in the source file." << endl;
      return false;
    }

    // execute the action       
    act = &m_actions[it1->second];
    cout << "-> Action: "; act->print();

    if (!act->is_executable(cs)) {
      //cout << endl << "STOP: Action " << *it 
	   //<< " cannot be executed." << endl;
		cout << endl << "STOP: Action ";
		act->print();
	    cout << " cannot be executed." << endl;
      return false;
    }

    cs = next_cstate(cs,act);
    
    // print the next state
    //cout << endl << "STEP " << n_count++ << endl;
  cout << "CState: ";
  cs->print();
    if (cs->goal_satisfied())
      cout << " (GOAL)";
    cout << endl;
  }
  
  return true;
}

Action* Planner::add_action(const string str)
{ 
  map<string,int>::iterator it;
  Action a(this,str);
  
  if (str.compare(0,9,"cpa_sense") && str.compare(0,5,"sense")) {  
	  it = m_action_map.find(str);
	  if (it == m_action_map.end()) {
		  // create a new entry	
			a.m_sensing = false;
			m_actions.push_back(a);
			m_action_map.insert(map<string,int>::value_type(str,m_actions.size()-1));
			return &(m_actions[m_actions.size()-1]);
	  }
	  return &(m_actions[it->second]);
  }

  //For sensing actions
  it = m_sensing_map.find(str);
  if (it == m_sensing_map.end()) {
	  // create a new entry			
	    a.m_sensing = true;
		m_sensings.push_back(a);
		m_sensing_map.insert(map<string,int>::value_type(str,m_sensings.size()-1));
		return &(m_sensings[m_sensings.size()-1]);
  } 
  return &(m_sensings[it->second]);

}

State* Planner::add_state(State* s)
{
  if (s->m_allocated) return s;
  pair<StateTable::iterator,bool> result;
  result = m_states[s->hash()].insert(s);
  if (!result.second) {
    // state exists    
	//if (s != *result.first)
		delete s;					
    return *result.first;
  }
  s->m_allocated = true;

  return s;
}

CSClause* Planner::add_clause(CSClause* s)
{
  if (s->m_allocated) return s;
  pair<ClauseTable::iterator,bool> result;
  result = m_clauses[s->hash()].insert(s);
  if (!result.second) {
    // state exists    
	//if (s != *result.first)				
		delete s;	//delete the duplicate later in allocate function
    return *result.first;
  }

  s->m_allocated = true;
  return s;
}


CState* Planner::next_cstate(CState* csin, const Action* act)
{   
  list<CState*> SF, SF1, *pS, *pS1, *pS2;
  list<CState*>::iterator itcs;

  EffectList::const_iterator it; 
  CState* cs;  

  SF.push_back(csin);

  pS = &SF;
  pS1 = &SF1;
   
  for (it = act->get_effects()->begin(); it != act->get_effects()->end(); ++it) {
	for (itcs = pS->begin(); itcs != pS->end(); itcs++)
		split_effect(*itcs, *it, pS1);
	pS2 = pS; pS = pS1; pS1 = pS2;	
	pS1->clear();
  }    
  //m_timer->start(ACTION_EXECUTE_TIMER);  
  action_execute(pS);  
  //m_timer->end(ACTION_EXECUTE_TIMER);
  
  //m_timer->start(MERGE_TIMER);  
  cs = or_merge(pS);
  //m_timer->end(MERGE_TIMER);  
  

  if (!cs->m_allocated) {
    //cs->Unit_Extract();    
	cs->allocate();	
	//cs->set_action(act);    
	//cs->set_previous_cstate(csin);
	//cs->set_plan_length(csin->get_plan_length()+1);
  }      
  return cs;
}

void Planner::action_execute(list<CState*> *SF) {
	Literals::const_iterator it_l, it_l1;
	Literal l;
	set<CSClause*>::iterator it_s, it_s1;
	int i;
	list<CState*>::iterator itcs;
	CState* cs;
	CSClause* s, *s1;
	set<CSClause*> S, S1;	
	for (itcs = SF->begin(); itcs != SF->end(); itcs++) {		
		cs = *itcs;
		//NEW FIXING		
		if 	(cs->m_allocated) 
			continue;
		//m_timer->start(SIMPLIFICATION_TIMER);	
		simplifyCNF(cs);
		//m_timer->end(SIMPLIFICATION_TIMER);
		Literals::iterator itl;
		Literal l1;
				
		for (it_l = cs->m_EF.begin(); it_l != cs->m_EF.end(); it_l++) {
			if (cs->m_L->m_L[*it_l]) {		
				continue;
			}	
			
			//Remove negation literal if exists
			NEGATE(*it_l,l);
			if (cs->m_L->m_L[l]) {				
				if (cs->is_inClause(l)) cout <<"Subsumed by Unit clause";
				if (cs->is_inClause(*it_l)) cout <<"Redundant by Unit clause";
				cs->m_L->m_L[l] = 0;
				cs->m_L->m_L[*it_l] = 1;
				continue;
			}			
			else { //Remove possible clauses containing
				if (cs->m_inS.size()) {
					while (cs->m_inS[l].size()) {
					//for (it_s = cs->m_inS[l].begin(); it_s != cs->m_inS[l].end(); it_s++) {
						s = *cs->m_inS[l].begin();
						for (it_l1 = s->m_literals.begin(); it_l1 != s->m_literals.end(); it_l1++)
							cs->m_inS[*it_l1].erase(s);											
						cs->m_clauses.erase(s);
						
						//If either l or -l doesn't have any clauses so the intersect set Clauses(l)*Clauses(-l) is empty
						if (cs->m_inS[*it_l].size()) {
							//s->m_literals.erase(l);					
							S.insert(s);
						}
						else { 
							if (!s->m_allocated)
								delete s;
						}
					}				
					cs->m_inS[l].clear();
				}
				else {
					for (it_s = cs->m_clauses.begin(); it_s != cs->m_clauses.end(); it_s++) {
						if ((*it_s)->m_literals.count(l))
							S.insert(*it_s);
						else if ((*it_s)->m_literals.count(*it_l))
							S1.insert(*it_s);
					}
				}
			}			
						
			cs->m_L->m_L[*it_l] = 1;
			
			if (cs->m_inS.size()) {
				while (cs->m_inS[*it_l].size()) {
				//for (it_s = cs->m_inS[*it_l].begin(); it_s != cs->m_inS[*it_l].end(); it_s++) {
					s = *cs->m_inS[*it_l].begin();					
					for (it_l1 = s->m_literals.begin(); it_l1 != s->m_literals.end(); it_l1++)						
						cs->m_inS[*it_l1].erase(s);
					
					cs->m_clauses.erase(s);
					
					//If either l or -l doesn't have any clauses so the intersect set Clauses(l)*Clauses(-l) is empty					
					if (S.size()) {
						//s->m_literals.erase(*it_l);							
						S1.insert(s);
					}
					else {			
						if (!s->m_allocated)
							delete s;						
					}
				}				
				cs->m_inS[*it_l].clear();
			}
			
			if (S.size() && S1.size()) {
				for (it_s = S.begin(); it_s != S.end(); it_s++) {
					(*it_s)->m_literals.erase(l);
					for (it_s1 = S1.begin(); it_s1 != S1.end(); it_s1++) {
						(*it_s1)->m_literals.erase(*it_l);
						s = new CSClause;
						s->m_literals = (*it_s)->m_literals;
						s->m_literals.insert((*it_s1)->m_literals.begin(), (*it_s1)->m_literals.end());
						if (s->m_literals.size() == 1) {
							cs->m_L->m_L[*s->m_literals.begin()] = 1;
							//m_timer->start(SIMPLIFICATION_TIMER);
							cs->UP(*s->m_literals.begin());
							//m_timer->end(SIMPLIFICATION_TIMER);							
						}
						else if (!s->is_consistent() || !cs->addClause(s, true)) 
								if (!s->m_allocated) 
									delete s;
					}
				}
			}
			
			if (!cs->m_inS.size()) {
				for (it_s = S.begin(); it_s != S.end(); it_s++) cs->m_clauses.erase(*it_s);
				for (it_s = S1.begin(); it_s != S1.end(); it_s++) cs->m_clauses.erase(*it_s);
			}
			
			for (it_s = S.begin(); it_s != S.end(); it_s++) {
				if (!(*it_s)->m_allocated)
					delete *it_s;				
			}
			S.clear();
			
			for (it_s1 = S1.begin(); it_s1 != S1.end(); it_s1++) {
				if (!(*it_s1)->m_allocated)
					delete *it_s1;				
			}
			S1.clear();			
		}

		//Treat nondeterministic effects

		for (it_l = cs->m_nondet_EF.begin(); it_l != cs->m_nondet_EF.end(); it_l++) {
			if (cs->m_L->m_L[*it_l]) {
				cs->m_L->m_L[*it_l]=0;
				continue;
			}
			
			//Remove negation literal if exists
			NEGATE(*it_l,l);
			if (cs->m_L->m_L[l]) {							
				cs->m_L->m_L[l] = 0;				
				continue;
			}						
			else { //Remove possible clauses containing
				if (cs->m_inS.size()) {
					while (cs->m_inS[l].size()) {
					//for (it_s = cs->m_inS[l].begin(); it_s != cs->m_inS[l].end(); it_s++) {
						s = *cs->m_inS[l].begin();
						for (it_l1 = s->m_literals.begin(); it_l1 != s->m_literals.end(); it_l1++)
							cs->m_inS[*it_l1].erase(s);											
						cs->m_clauses.erase(s);
						
						//If either l or -l doesn't have any clauses so the intersect set Clauses(l)*Clauses(-l) is empty
						if (cs->m_inS[*it_l].size()) {
							//s->m_literals.erase(l);					
							S.insert(s);
						}
						else { 
							if (!s->m_allocated)
								delete s;
						}
					}				
					cs->m_inS[l].clear();
				}
				else {
					for (it_s = cs->m_clauses.begin(); it_s != cs->m_clauses.end(); it_s++) {
						if ((*it_s)->m_literals.count(l))
							S.insert(*it_s);
						else if ((*it_s)->m_literals.count(*it_l))
							S1.insert(*it_s);
					}
				}
			}			
			
			if (cs->m_inS.size()) {
				while (cs->m_inS[*it_l].size()) {
				//for (it_s = cs->m_inS[*it_l].begin(); it_s != cs->m_inS[*it_l].end(); it_s++) {
					s = *cs->m_inS[*it_l].begin();					
					for (it_l1 = s->m_literals.begin(); it_l1 != s->m_literals.end(); it_l1++)						
						cs->m_inS[*it_l1].erase(s);
					
					cs->m_clauses.erase(s);
					
					//If either l or -l doesn't have any clauses so the intersect set Clauses(l)*Clauses(-l) is empty					
					if (S.size()) {
						//s->m_literals.erase(*it_l);							
						S1.insert(s);
					}
					else {			
						if (!s->m_allocated)
							delete s;						
					}
				}				
				cs->m_inS[*it_l].clear();
			}
			
			if (S.size() && S1.size()) {
				for (it_s = S.begin(); it_s != S.end(); it_s++) {
					(*it_s)->m_literals.erase(l);
					for (it_s1 = S1.begin(); it_s1 != S1.end(); it_s1++) {
						(*it_s1)->m_literals.erase(*it_l);
						s = new CSClause;
						s->m_literals = (*it_s)->m_literals;
						s->m_literals.insert((*it_s1)->m_literals.begin(), (*it_s1)->m_literals.end());
						if (s->m_literals.size() == 1) {
							cs->m_L->m_L[*s->m_literals.begin()] = 1;
							//m_timer->start(SIMPLIFICATION_TIMER);
							cs->UP(*s->m_literals.begin());
							//m_timer->end(SIMPLIFICATION_TIMER);
						}
						else if (!s->is_consistent() || !cs->addClause(s, true)) 
								if (!s->m_allocated) 
									delete s;
					}
				}
			}
			
			if (!cs->m_inS.size()) {
				for (it_s = S.begin(); it_s != S.end(); it_s++) cs->m_clauses.erase(*it_s);
				for (it_s = S1.begin(); it_s != S1.end(); it_s++) cs->m_clauses.erase(*it_s);
			}
			
			for (it_s = S.begin(); it_s != S.end(); it_s++) {
				if (!(*it_s)->m_allocated)
					delete *it_s;				
			}
			S.clear();
			
			for (it_s1 = S1.begin(); it_s1 != S1.end(); it_s1++) {
				if (!(*it_s1)->m_allocated)
					delete *it_s1;				
			}
			S1.clear();			
		}		
	}	
}


CState* Planner::or_merge(list<CState*> *SF) {
	CState *cs;
	int i;	
	cs = *SF->begin();
	//cout << "0"; fflush(stdout);
	if (SF->size()>1) {	
		//cout << "1"; fflush(stdout);
		list<CState*>::iterator itcs;
		CState *cs1, *cs_new;
		cs_new = new CState;
		cs_new->m_planner = this;
		cs_new->m_L = new State(this);
		itcs = SF->begin();
		//cout << "2"; fflush(stdout);
		
		for ( ++itcs; itcs != SF->end(); itcs++) {
			//cout << "3"; fflush(stdout);
			cs1 = *itcs;			
			cs->or_merge(cs1,cs_new);// cs_new = cs OR cs1			
			//cout << "4"; fflush(stdout);
			cs1->clear();
			if (!cs1->m_L->m_allocated)
				delete cs1->m_L;
			delete cs1;
			//cout << "5"; fflush(stdout);
			cs->clear();			
			cs1 = cs;
			cs = cs_new;
			cs_new = cs1;
			//cout << "6"; fflush(stdout);
		}
		//cout << "7"; fflush(stdout);
		if (!cs_new->m_L->m_allocated)
			delete cs_new->m_L;
		delete cs_new;
		//cout << "8"; fflush(stdout);
	}
	
	if (!cs->m_allocated) {		
		//m_timer->start(SIMPLIFICATION_TIMER);
		simplifyCNF(cs);
		//m_timer->end(SIMPLIFICATION_TIMER);
		cs->m_EF.clear();
		cs->m_nondet_EF.clear();
	}

	//cout << "9"; fflush(stdout);
	return cs;
}


void Planner::split_effect(CState* cs,const Effect& ef, list<CState*>* SF1) {
	
	const Literals *bd;
	const Literals *hd;
		
	bd = ef.get_body();
    hd = ef.get_head();	
	int sat_v;

	//m_timer->start(SAT_CHECKING_TIMER);
	sat_v = sat_value(cs,ef);
	//m_timer->end(SAT_CHECKING_TIMER);

	switch (sat_v) {
		case 0: SF1->push_back(cs); 
			    break;
				
		case 1: if (!cs->m_allocated) {
					if (!ef.nondet())
						cs->m_EF.insert(hd->begin(), hd->end());
					else
						cs->m_nondet_EF.insert(hd->begin(), hd->end());
					SF1->push_back(cs);
				}
				else {//cout << " Test 1 "; fflush(stdout);
					CState* cs1 = new CState; 
					cs1->copy(cs);//cout << " Test 2 "; fflush(stdout);
					if (!ef.nondet())
						cs1->m_EF.insert(hd->begin(), hd->end());
					else
						cs1->m_nondet_EF.insert(hd->begin(), hd->end());
					SF1->push_back(cs1);					
				}
				break;

		case 2: if (!cs->m_allocated) {
					if (!ef.nondet())
						cs->m_EF.insert(hd->begin(), hd->end());
					else
						cs->m_nondet_EF.insert(hd->begin(), hd->end());
					SF1->push_back(cs);
				}
				else {//cout << " Test 3 "; fflush(stdout);
					CState* cs1 = new CState; 
					cs1->copy(cs);//cout << " Test 4 "; fflush(stdout);
					if (!ef.nondet())
						cs1->m_EF.insert(hd->begin(), hd->end());
					else
						cs1->m_nondet_EF.insert(hd->begin(), hd->end());
					SF1->push_back(cs1);
				}
				break;

		case 3: CState* cs1 = new CState; 
				cs1->copy(cs); 
				if (!cs->m_allocated) { 
					if (!ef.nondet())
						cs->m_EF.insert(hd->begin(), hd->end()); 
					else
						cs->m_nondet_EF.insert(hd->begin(), hd->end());
					//m_timer->start(SIMPLIFICATION_TIMER);
					cs->UP(*bd); 
					//m_timer->end(SIMPLIFICATION_TIMER);
					SF1->push_back(cs);
				}
				else {
					CState* cs2 = new CState; 
					cs2->copy(cs); 
					if (!ef.nondet())
						cs2->m_EF.insert(hd->begin(), hd->end());
					else
						cs2->m_nondet_EF.insert(hd->begin(), hd->end());
					//m_timer->start(SIMPLIFICATION_TIMER);
					cs2->UP(*bd);
					//m_timer->end(SIMPLIFICATION_TIMER);
					SF1->push_back(cs2);
				}				
				cs1->and_merge_negation(bd); 
				SF1->push_back(cs1);
				break;
	}		
	
	//cout << "e"; fflush(stdout);
	return;
}


// 0- cs entailes -bd, 1- cs entails bd by unit literals, 2- cs entails bd by sat solver, 3-unknown
int Planner::sat_value(CState* cs, const Effect& ef) {
	Literals::const_iterator it_l;
	Literal l;
	set<State*>::const_iterator its;
	bool pass;
	const Literals *bd;
	const Literals *hd;
	
	bd = ef.get_body();
    hd = ef.get_head();
	
	//Check for F obviously entails -bd
	for (it_l = bd->begin(); it_l!= bd->end();it_l++) {
		NEGATE(*it_l,l);
		if (cs->m_L->m_L[l]) {			
			return 0;
		}
	}	
	//If the state already includes the head of effect so we don't need to do anything
	
	if (cs->includes(hd)) {	
		return 0;
	}
	
	//if (cs->will_include(hd)) {		
	//	cout << "WILL";
	//	return 0;
	//}
	
	if (cs->includes(bd)) {
		return 1;
	}	
	
	//Check for F & -bd is unsatisfiable or F entails bd
	pass = true;
	for (it_l = bd->begin(); it_l!=bd->end();it_l++) {
		if (!cs->is_inClause(*it_l) && !cs->m_L->m_L[*it_l]) {
			pass = false;
			break;
		}
	}
	//cout << "1"; fflush(stdout);
	if (pass) {		
		bool sol_neg = cs->solvable_assume_negation(bd);		
		//if(!cs->solvable_assume_negation(bd)) {
		if(!sol_neg) {
			//cout << " UNSAT UNIT "<<endl; fflush(stdout);
			//cs->Unit_Extract();
			//cout << " UNSAT UNIT CHECK "<<endl; fflush(stdout);
			//m_timer->start(SIMPLIFICATION_TIMER);
			cs->UP(*bd);
			//m_timer->end(SIMPLIFICATION_TIMER);
			return 2;
		}	
	}

	//Check for F & bd is unsatisfiable or F entails -bd	
	pass = true;
	for (it_l = bd->begin(); it_l!=bd->end();it_l++) {
		if (!cs->m_L->m_L[*it_l]) {
			NEGATE(*it_l,l);
			if (cs->is_inClause(l)) {
				pass = false;
				break;
			}
		}
	}

	if (!pass) {		
		bool sol = cs->solvable_assume(bd);		
		//if(!sol && !cs->m_allocated){		
		if(!sol){				
			//m_timer->start(SIMPLIFICATION_TIMER);
			simplifyCNF(cs);
			//cs->UP(); ????
			//m_timer->end(SIMPLIFICATION_TIMER);			
			return 0;
		}
	}

	//++m_unknown_antecedents_sum;

	return 3;
}


Literals Planner::negate(const Literals& x) const
{
  Literals::const_iterator it;
  Literals y; // output literals
  Literal l;
  
  for (it = x.begin(); it != x.end(); ++it) {
    NEGATE(*it,l);
    y.insert(l);
  }
  return y;
}

CState* Planner::first_unexpanded()
{	
	CState* cs;	

	while (m_queue.size() && (m_queue.top()->m_status != UNEXPANDED)) {
		if (m_queue.top()->m_status == ISOLATED)
			m_queue.top()->m_status = OUT_ISOLATED;
		m_queue.pop();
	}

	if (m_queue.size()) {
		cs = m_queue.top();
		m_queue.pop();
		return cs;
	}

	return NULL;
}


bool Planner::search()
{
  CState *cs;
  set<CSClause*>::iterator its;
  	
  if (m_init_cstate.goal_satisfied()) {
    cout << "The initial state satisies the goal " << endl;
	return true;
  }  
  
  m_init_cstate.m_status = UNEXPANDED;
  m_expanded = 0;
  //m_expanded_clauses =0;
  //m_expanded_sum_clause_size =0;
  //m_unknown_antecedents_sum = 0;
  //m_exec_actions = 0;
  m_queue.push(&m_init_cstate);
  m_cstates[m_init_cstate.hash()].insert(&m_init_cstate);    

  while ((cs = first_unexpanded())) {	  
	  m_expanded++;
	  //m_expanded_clauses += cs->m_clauses.size();
	  //for (its = cs->m_clauses.begin(); its != cs->m_clauses.end(); its++) {
		//  m_expanded_sum_clause_size += (*its)->m_literals.size();
	  //}
	  if (expand(cs))
		  break;	  
  }      
  
  if (m_init_cstate.m_status != GOAL_REACHABLE) {
    cout << "No plan was found." << endl;
	return false;
  }
  
  cout << endl << "  Found a plan: " <<endl;  
  m_depth = 0;
  //m_init_cstate.solution_not_print(this,0);  
  m_init_cstate.solution_print(this,0);  
  cout <<"Size of solution: " << m_commandNo.size() <<"\t Depth: " << m_depth << endl;

  return true;
}

//Return true if search is terminated
bool Planner::expand(CState *cs)
{
  CState *cs1, *cs2;
  const Action *act;
  ActionList::const_iterator it;    
  CStateTable::iterator it2;
  static int maxhvalue =-1; 
    
  //Usual actions

  //cout <<" Expanding cs: "<<cs <<", "; cs->print(); cout <<"cs1 = "<< cs1 << endl;

  cs->inClauses();

  for (it = m_actions.begin(); it != m_actions.end(); it++) {	  
	  act = &(*it);    		
	  //m_timer->start(SAT_CHECKING_TIMER);
	  if (!act->is_executable(cs)) {		  
		  //m_timer->end(SAT_CHECKING_TIMER);	
		  continue;	  
	  }
  	  else
	 // m_timer->end(SAT_CHECKING_TIMER);	  
	  cs1 = next_cstate(cs,act);
	  
	  if (cs1 == cs)
		  continue;

	  ++m_exec_actions;
	  
	  new_action_transition(cs, act, cs1);	  	  	  	  
		
	  //Terminate the search when the initial state reaches the goal or a dead-end	  
	  if (m_init_cstate.m_status == GOAL_REACHABLE) {
		  cs->m_inS.clear();
		  return true;
	  }

	  if (cs->m_status == GOAL_REACHABLE) {
		  cs->m_inS.clear();
		  return false;	  
	  }
  }
    

  //Sensing actions
  for (it = m_sensings.begin(); it != m_sensings.end(); it++) {
	  act = &(*it); 
	  //m_timer->start(SAT_CHECKING_TIMER);
	  if (!act->is_executable(cs) || cs->decides(act->lit())) {		  
		  //m_timer->end(SAT_CHECKING_TIMER);
		  continue;	  
	  }
	  else
		//m_timer->end(SAT_CHECKING_TIMER);
	  	  
	  cs1 = new CState;
	  cs2 = new CState;
	  cs->split(cs1, cs2, act->lit());	  	
	  new_sensing_transition(cs, act, cs1, cs2);	  

	  if (m_init_cstate.m_status == GOAL_REACHABLE) {
		  cs->m_inS.clear();
		  return true;
	  }
	  
	  if (cs->m_status == GOAL_REACHABLE) {
		  cs->m_inS.clear();
		  return false;
	  }
  }  
  
  cs->m_inS.clear();

  if (!cs->m_successors.size()) {
	  cs->m_status = DEAD;
	  cout << "DEAD " << cs << endl;	  
	  dead_propagate(cs);	  
	  if (m_init_cstate.m_status == DEAD)
		  return true;
  }  
  return false;
}

//act: cs -> cs1
bool Planner::new_action_transition(CState* cs, const Action *act, CState* cs1)
{	
	CStateTable::iterator it;	
	int hash;	

	hash = cs1->hash();		
	it = m_cstates[hash].find(cs1);	 	
	  
	if (it == m_cstates[hash].end()) {
		cs1->hvalue();
		//m_timer->start(SAT_CHECKING_TIMER);
		if (cs1->goal_satisfied()) {			
			//m_timer->end(SAT_CHECKING_TIMER);
			cs1->m_status = GOAL_REACHABLE;			
			cs->m_status = GOAL_REACHABLE;
			goal_propagate(cs, cs1);				
		}		
		else {			
			//m_timer->end(SAT_CHECKING_TIMER);
			m_queue.push(cs1);
			//m_queue1.push(cs1);			
		}		
		m_cstates[hash].insert(cs1);		
	}
	else {
		delete cs1;
		cs1 = *it;
		//Dead-end
		if (cs1 == cs || cs1->m_status == DEAD)
			return false;			

		if (cs1->m_status == GOAL_REACHABLE) {			
			cs->m_status = GOAL_REACHABLE;			
			goal_propagate(cs,cs1);		
		}		
		else {// cs1 is EXPANDED			
			reconnection_propagate(cs1);			
		}
	}		
	cs->m_successors.push_back(cs1);
	cs1->m_in_transitions.insert(pair<CState*,const Action*>(cs,act));	

	return true;
}

bool Planner::new_sensing_transition(CState* cs, const Action *act, CState* cs1, CState* cs2)
{	
	CStateTable::iterator it1, it2;	
	int hash1, hash2;    
	hash1 = cs1->hash();
	it1 = m_cstates[hash1].find(cs1);
	
	hash2 = cs2->hash();
	it2 = m_cstates[hash2].find(cs2);
		
	if ((it1 != m_cstates[hash1].end() && (*it1)->m_status == DEAD) || 
		(it2 != m_cstates[hash2].end() && (*it2)->m_status == DEAD )) {
			delete cs1;
			delete cs2;
			return false;
	}

	
	if ((it1 != m_cstates[hash1].end() && ((*it1)->m_status == ISOLATED || (*it1)->m_status == OUT_ISOLATED)) &&
		(it2 != m_cstates[hash2].end() && ((*it2)->m_status == ISOLATED || (*it2)->m_status == OUT_ISOLATED))) {
			delete cs1;
			delete cs2;
			cout << "ISOLATION, IGNORE SENSING";
			return false;
	}
	

	//Treat cs1	  
	if (it1 == m_cstates[hash1].end()) {
		//m_timer->start(SAT_CHECKING_TIMER);
		if (cs1->goal_satisfied())			
			cs1->m_status = GOAL_REACHABLE;
		//m_timer->end(SAT_CHECKING_TIMER);
		m_queue.push(cs1);		
		//m_queue1.push(cs1);
		m_cstates[hash1].insert(cs1);		  
	}
	else {
		delete cs1;
		cs1 = *it1;		
		cs1->m_active_in_transitions++;				
		if (cs1->m_status == ISOLATED) {
			cs1->m_status = UNEXPANDED;
		}
		else if (cs1->m_status == OUT_ISOLATED) {
			cs1->m_status = UNEXPANDED;
			m_queue.push(cs1);
			//m_queue1.push(cs1);
		}		
	}
	cs->m_successors.push_back(cs1);
	cs1->m_in_transitions.insert(pair<CState*,const Action*>(cs,act));	
	
	//Treat cs2
	if (it2 == m_cstates[hash2].end()) {		
		//m_timer->start(SAT_CHECKING_TIMER);
		if (cs2->goal_satisfied()){			
			//m_timer->end(SAT_CHECKING_TIMER);
			cs2->m_status = GOAL_REACHABLE;	
			if (cs1->m_status == GOAL_REACHABLE) {
				cs->m_status = GOAL_REACHABLE;
				goal_propagate(cs,cs1,cs2);
			}
		}		
		else
			//m_timer->end(SAT_CHECKING_TIMER);
		m_queue.push(cs2);
		//m_queue1.push(cs2);
		m_cstates[hash2].insert(cs2);
	}
	else {
		delete cs2;
		cs2 = *it2;		
		cs2->m_active_in_transitions++;
		if (cs2->m_status == GOAL_REACHABLE) {			
			if (cs1->m_status == GOAL_REACHABLE) {
				cs->m_status = GOAL_REACHABLE;
				goal_propagate(cs,cs1,cs2);
			}
		}
		else if (cs2->m_status == ISOLATED) {
			cs2->m_status = UNEXPANDED;
		}
		else if (cs2->m_status == OUT_ISOLATED) {
			cs2->m_status = UNEXPANDED;			
			m_queue.push(cs2);
			//m_queue1.push(cs2);
		}
		else {// cs2 is EXPANDED
			reconnection_propagate(cs2);
			if (cs1->m_status == EXPANDED) {
				reconnection_propagate(cs1);
			}
		}
	}
	cs->m_successors.push_back(cs2);
	cs2->m_in_transitions.insert(pair<CState*,const Action*>(cs,act));		
	return true;
}

void Planner::goal_propagate(CState* cs, CState* suc_goal)
{
	vector<CState*>::iterator itcs;	
	map<CState*,const Action*>::iterator it;
		
	for (itcs = cs->m_successors.begin(); itcs != cs->m_successors.end(); itcs++) {
		if (*itcs != suc_goal) {
			isolation_propagate(*itcs);					
			(*itcs)->m_in_transitions.erase(cs);
		}
	}

	//Only keep the outgoing transition(s) to the goal node(s)
	cs->m_successors.clear();
	cs->m_successors.push_back(suc_goal);
	
	for (it = cs->m_in_transitions.begin(); it != cs->m_in_transitions.end(); it++) {
		if (it->second->m_sensing) {
			//Two complement successors should be next to one another
			for (itcs = it->first->m_successors.begin(); itcs != it->first->m_successors.end(); itcs++) {
				if ((*itcs)->is_successor(it->first, it->second))
					break;
			}
			//The complement of cs should be next to cs
					
			if (*itcs == cs) {
				itcs++;
				if ((*itcs)->m_status == GOAL_REACHABLE) {
					it->first->m_status = GOAL_REACHABLE;
					goal_propagate(it->first, cs, *itcs);
				}
			}
			else {								
				if ((*itcs)->m_status == GOAL_REACHABLE) {
					it->first->m_status = GOAL_REACHABLE;
					goal_propagate(it->first, *itcs, cs);
				}				
			}					
		}
		else {
			it->first->m_status = GOAL_REACHABLE;
			goal_propagate(it->first, cs);
		}
	}
}

void Planner::goal_propagate(CState* cs, CState* suc_goal1, CState* suc_goal2)
{
	vector<CState*>::iterator itcs;	
	map<CState*,const Action*>::iterator it;
		
	for (itcs = cs->m_successors.begin(); itcs != cs->m_successors.end(); itcs++) {
		if (*itcs != suc_goal1 && *itcs != suc_goal2) {
			isolation_propagate(*itcs);					
			(*itcs)->m_in_transitions.erase(cs);
		}
	}

	//Only keep the outgoing transition(s) to the goal node(s)
	cs->m_successors.clear();
	cs->m_successors.push_back(suc_goal1);
	cs->m_successors.push_back(suc_goal2);
	
	for (it = cs->m_in_transitions.begin(); it != cs->m_in_transitions.end(); it++) {
		if (it->second->m_sensing) {
			//Two complement successors should be next to one another
			for (itcs = it->first->m_successors.begin(); itcs != it->first->m_successors.end(); itcs++) {
				if ((*itcs)->is_successor(it->first, it->second))
					break;
			}
			//The complement of cs should be next to cs
					
			if (*itcs == cs) {
				itcs++;
				if ((*itcs)->m_status == GOAL_REACHABLE) {
					it->first->m_status = GOAL_REACHABLE;
					goal_propagate(it->first, cs, *itcs);
				}
			}
			else {								
				if ((*itcs)->m_status == GOAL_REACHABLE) {
					it->first->m_status = GOAL_REACHABLE;
					goal_propagate(it->first, *itcs, cs);
				}				
			}					
		}
		else {
			it->first->m_status = GOAL_REACHABLE;
			goal_propagate(it->first, cs);
		}
	}
}

void Planner::dead_propagate(CState* cs)
{	
	map<CState*,const Action*>::iterator it;
	vector<CState*>::iterator itcs;			
	while (cs->m_in_transitions.size()) {
		it = cs->m_in_transitions.begin();
		if (it->second->m_sensing) {			
			//Two complement successors should be next to one another
			for (itcs = it->first->m_successors.begin(); itcs != it->first->m_successors.end(); itcs++) {				
				if ((*itcs)->is_successor(it->first, it->second))
					break;
			}
			//The complement of cs should be next to cs
					
			if (*itcs == cs) {
				itcs++;
				isolation_propagate(*itcs);				
				(*itcs)->m_in_transitions.erase(it->first);								
				it->first->m_successors.erase(itcs-1, itcs+1);				
			}
			else {										
				isolation_propagate(*itcs);				
				(*itcs)->m_in_transitions.erase(it->first);
				it->first->m_successors.erase(itcs, itcs+2);				
			}			
			if (!it->first->m_successors.size()) {
				it->first->m_status = DEAD;
				dead_propagate(it->first);
			}
		}
		else {
			for (itcs = it->first->m_successors.begin(); itcs != it->first->m_successors.end(); itcs++) {
				if (*itcs == cs)
					break;
			}
			it->first->m_successors.erase(itcs);
			if (!it->first->m_successors.size()) {
				it->first->m_status = DEAD;
				dead_propagate(it->first);
			}
		}
		cs->m_in_transitions.erase(it);
	}	
}

void Planner::isolation_propagate(CState* cs)
{
	//return;
	vector<CState*>::iterator itcs;			
	cs->m_active_in_transitions--;
	if (cs->m_active_in_transitions || cs->m_status == GOAL_REACHABLE)
		return;	

	if (cs->m_status == UNEXPANDED) {
		cs->m_status = ISOLATED;
		return;
	}	

	for (itcs = cs->m_successors.begin(); itcs != cs->m_successors.end(); itcs++) {		
		isolation_propagate(*itcs);
	}
}

void Planner::reconnection_propagate(CState* cs)
{
	//return;
	vector<CState*>::iterator itcs;
	cs->m_active_in_transitions++;
	if (cs->m_active_in_transitions > 1 || cs->m_status == GOAL_REACHABLE)
		return;
	if (cs->m_status == ISOLATED) 
		cs->m_status = UNEXPANDED;
	else if (cs->m_status == OUT_ISOLATED) {
		cs->m_status = UNEXPANDED;		
		m_queue.push(cs);
		//m_queue1.push(cs);
	}
	else {// cs1 is EXPANDED
		for (itcs = cs->m_successors.begin(); itcs != cs->m_successors.end(); itcs++)		
			reconnection_propagate(*itcs);
	}
}